import os
import random

# Fix for OpenMP library conflict on Windows
os.environ["KMP_DUPLICATE_LIB_OK"] = "TRUE"

import torch

from pytorch3dunet.unet3d.config import copy_config, load_config
from pytorch3dunet.unet3d.trainer import create_trainer
from pytorch3dunet.unet3d.utils import get_logger

logger = get_logger("TrainingSetup")


def main():
    """Main entry point for training 3D U-Net models.

    Loads configuration from command line arguments, sets random seeds if specified,
    creates a trainer instance, and starts the training process.
    """
    # Load and log experiment configuration
    config, config_path = load_config()
    logger.info(config)

    manual_seed = config.get("manual_seed", None)
    if manual_seed is not None:
        logger.info(f"Seed the RNG for all devices with {manual_seed}")
        logger.warning("Using CuDNN deterministic setting. This may slow down the training!")
        random.seed(manual_seed)
        torch.manual_seed(manual_seed)
        # see https://pytorch.org/docs/stable/notes/randomness.html
        torch.backends.cudnn.deterministic = True

    # Create trainer
    trainer = create_trainer(config)
    # Copy config file
    copy_config(config, config_path)
    # Start training
    trainer.fit()


if __name__ == "__main__":
    main()



from torch import nn

from pytorch3dunet.unet3d.buildingblocks import DoubleConv, ResNetBlock, ResNetBlockSE, create_decoders, create_encoders
from pytorch3dunet.unet3d.utils import get_class, number_of_features_per_level


class AbstractUNet(nn.Module):
    """Base class for standard and residual UNet.

    Args:
        in_channels: Number of input channels.
        out_channels: Number of output segmentation masks. Note that the number of out_channels might correspond
            to either different semantic classes or to different binary segmentation mask. It's up to the user
            of the class to interpret the out_channels and use the proper loss criterion during training
            (i.e. CrossEntropyLoss for multi-class or BCEWithLogitsLoss for two-class respectively).
        final_sigmoid: If True apply element-wise nn.Sigmoid after the final 1x1 convolution, otherwise apply
            nn.Softmax. In effect only if `self.training == False`, i.e. during validation/testing.
        basic_module: Basic model for the encoder/decoder (DoubleConv, ResNetBlock, etc.).
        f_maps: Number of feature maps at each level of the encoder. If it's an integer, the number of feature
            maps is given by the geometric progression: f_maps ^ k, k=1,2,3,4. Default: 64.
        layer_order: Determines the order of layers in `SingleConv` module. E.g. 'crg' stands for
            GroupNorm3d+Conv3d+ReLU. See `SingleConv` for more info. Default: 'gcr'.
        num_groups: Number of groups for the GroupNorm. Default: 8.
        num_levels: Number of levels in the encoder/decoder path (applied only if f_maps is an int). Default: 4.
        is_segmentation: If True and the model is in eval mode, Sigmoid/Softmax normalization is applied after
            the final convolution. If False (regression problem) the normalization layer is skipped. Default: True.
        conv_kernel_size: Size of the convolving kernel in the basic_module. Default: 3.
        pool_kernel_size: The size of the pooling window. Default: 2.
        conv_padding: Zero-padding added to all three sides of the input. Default: 1.
        conv_upscale: Number of the convolution to upscale in encoder if DoubleConv. Default: 2.
        upsample: Algorithm used for decoder upsampling. Options are 'nearest', 'linear', 'bilinear', 'trilinear',
            'area' (InterpolateUpsampling), 'deconv' (TransposeConvUpsampling), or None (no upsampling).
            Default: 'default' (chooses automatically).
        dropout_prob: Dropout probability. Default: 0.1.
        is3d: If True the model is 3D, otherwise 2D. Default: True.
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        final_sigmoid,
        basic_module,
        f_maps=64,
        layer_order="gcr",
        num_groups=8,
        num_levels=4,
        is_segmentation=True,
        conv_kernel_size=3,
        pool_kernel_size=2,
        conv_padding=1,
        conv_upscale=2,
        upsample="default",
        dropout_prob=0.1,
        is3d=True,
    ):
        super().__init__()

        if isinstance(f_maps, int):
            f_maps = number_of_features_per_level(f_maps, num_levels=num_levels)

        assert isinstance(f_maps, list) or isinstance(f_maps, tuple)
        assert len(f_maps) > 1, "Required at least 2 levels in the U-Net"
        if "g" in layer_order:
            assert num_groups is not None, "num_groups must be specified if GroupNorm is used"

        # create encoder path
        self.encoders = create_encoders(
            in_channels,
            f_maps,
            basic_module,
            conv_kernel_size,
            conv_padding,
            conv_upscale,
            dropout_prob,
            layer_order,
            num_groups,
            pool_kernel_size,
            is3d,
        )

        # create decoder path
        self.decoders = create_decoders(
            f_maps, basic_module, conv_kernel_size, conv_padding, layer_order, num_groups, upsample, dropout_prob, is3d
        )

        # in the last layer a 1×1 convolution reduces the number of output channels to the number of labels
        if is3d:
            self.final_conv = nn.Conv3d(f_maps[0], out_channels, 1)
        else:
            self.final_conv = nn.Conv2d(f_maps[0], out_channels, 1)

        if is_segmentation:
            # semantic segmentation problem
            if final_sigmoid:
                self.final_activation = nn.Sigmoid()
            else:
                self.final_activation = nn.Softmax(dim=1)
        else:
            # regression problem
            self.final_activation = None

    def forward(self, x, return_logits=False):
        """
        Forward pass through the network.

        Args:
            x (torch.Tensor): Input tensor of shape (N, C, D, H, W) for 3D or (N, C, H, W) for 2D,
                              where N is the batch size, C is the number of channels,
                              D is the depth, H is the height, and W is the width.
            return_logits (bool): If True, returns both the output and the logits.
                                  If False, returns only the output. Default is False.

        Returns:
            torch.Tensor: The output tensor after passing through the network.
                          If return_logits is True, returns a tuple of (output, logits).
        """
        output, logits = self._forward_logits(x)
        if return_logits:
            return output, logits
        return output

    def _forward_logits(self, x):
        # encoder part
        encoders_features = []
        for encoder in self.encoders:
            x = encoder(x)
            # reverse the encoder outputs to be aligned with the decoder
            encoders_features.insert(0, x)

        # remove the last encoder's output from the list
        # !!remember: it's the 1st in the list
        encoders_features = encoders_features[1:]

        # decoder part
        for decoder, encoder_features in zip(self.decoders, encoders_features, strict=False):
            # pass the output from the corresponding encoder and the output
            # of the previous decoder
            x = decoder(encoder_features, x)

        x = self.final_conv(x)

        if self.final_activation is not None:
            # compute final activation
            out = self.final_activation(x)
            # return both probabilities and logits
            return out, x

        return x, x


class UNet3D(AbstractUNet):
    """3D U-Net model from "3D U-Net: Learning Dense Volumetric Segmentation from Sparse Annotation".

    Uses `DoubleConv` as a basic_module and nearest neighbor upsampling in the decoder.
    Reference: https://arxiv.org/pdf/1606.06650.pdf
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        final_sigmoid=True,
        f_maps=64,
        layer_order="gcr",
        num_groups=8,
        num_levels=4,
        is_segmentation=True,
        conv_padding=1,
        conv_upscale=2,
        upsample="default",
        dropout_prob=0.1,
        **kwargs,
    ):
        super().__init__(
            in_channels=in_channels,
            out_channels=out_channels,
            final_sigmoid=final_sigmoid,
            basic_module=DoubleConv,
            f_maps=f_maps,
            layer_order=layer_order,
            num_groups=num_groups,
            num_levels=num_levels,
            is_segmentation=is_segmentation,
            conv_padding=conv_padding,
            conv_upscale=conv_upscale,
            upsample=upsample,
            dropout_prob=dropout_prob,
            is3d=True,
        )


class ResidualUNet3D(AbstractUNet):
    """Residual 3D U-Net model implementation.

    Uses ResNetBlock as a basic building block, summation joining instead of concatenation joining,
    and transposed convolutions for upsampling (watch out for block artifacts). Since the model
    effectively becomes a residual net, in theory it allows for deeper UNet.

    Reference: https://arxiv.org/pdf/1706.00120.pdf
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        final_sigmoid=True,
        f_maps=64,
        layer_order="gcr",
        num_groups=8,
        num_levels=5,
        is_segmentation=True,
        conv_padding=1,
        conv_upscale=2,
        upsample="default",
        dropout_prob=0.1,
        **kwargs,
    ):
        super().__init__(
            in_channels=in_channels,
            out_channels=out_channels,
            final_sigmoid=final_sigmoid,
            basic_module=ResNetBlock,
            f_maps=f_maps,
            layer_order=layer_order,
            num_groups=num_groups,
            num_levels=num_levels,
            is_segmentation=is_segmentation,
            conv_padding=conv_padding,
            conv_upscale=conv_upscale,
            upsample=upsample,
            dropout_prob=dropout_prob,
            is3d=True,
        )


class ResidualUNetSE3D(AbstractUNet):
    """Residual 3D U-Net model implementation with Squeeze and Excitation blocks.

    Uses ResNetBlockSE as a basic building block, summation joining instead of concatenation joining,
    and transposed convolutions for upsampling (watch out for block artifacts). Since the model
    effectively becomes a residual net, in theory it allows for deeper UNet.

    Reference: https://arxiv.org/pdf/1706.00120.pdf
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        final_sigmoid=True,
        f_maps=64,
        layer_order="gcr",
        num_groups=8,
        num_levels=5,
        is_segmentation=True,
        conv_padding=1,
        conv_upscale=2,
        upsample="default",
        dropout_prob=0.1,
        **kwargs,
    ):
        super().__init__(
            in_channels=in_channels,
            out_channels=out_channels,
            final_sigmoid=final_sigmoid,
            basic_module=ResNetBlockSE,
            f_maps=f_maps,
            layer_order=layer_order,
            num_groups=num_groups,
            num_levels=num_levels,
            is_segmentation=is_segmentation,
            conv_padding=conv_padding,
            conv_upscale=conv_upscale,
            upsample=upsample,
            dropout_prob=dropout_prob,
            is3d=True,
        )


class UNet2D(AbstractUNet):
    """2D U-Net model from "U-Net: Convolutional Networks for Biomedical Image Segmentation".

    Reference: https://arxiv.org/abs/1505.04597
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        final_sigmoid=True,
        f_maps=64,
        layer_order="gcr",
        num_groups=8,
        num_levels=4,
        is_segmentation=True,
        conv_padding=1,
        conv_upscale=2,
        upsample="default",
        dropout_prob=0.1,
        **kwargs,
    ):
        super().__init__(
            in_channels=in_channels,
            out_channels=out_channels,
            final_sigmoid=final_sigmoid,
            basic_module=DoubleConv,
            f_maps=f_maps,
            layer_order=layer_order,
            num_groups=num_groups,
            num_levels=num_levels,
            is_segmentation=is_segmentation,
            conv_padding=conv_padding,
            conv_upscale=conv_upscale,
            upsample=upsample,
            dropout_prob=dropout_prob,
            is3d=False,
        )


class ResidualUNet2D(AbstractUNet):
    """Residual 2D U-Net model implementation.

    Reference: https://arxiv.org/pdf/1706.00120.pdf
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        final_sigmoid=True,
        f_maps=64,
        layer_order="gcr",
        num_groups=8,
        num_levels=5,
        is_segmentation=True,
        conv_padding=1,
        conv_upscale=2,
        upsample="default",
        dropout_prob=0.1,
        **kwargs,
    ):
        super().__init__(
            in_channels=in_channels,
            out_channels=out_channels,
            final_sigmoid=final_sigmoid,
            basic_module=ResNetBlock,
            f_maps=f_maps,
            layer_order=layer_order,
            num_groups=num_groups,
            num_levels=num_levels,
            is_segmentation=is_segmentation,
            conv_padding=conv_padding,
            conv_upscale=conv_upscale,
            upsample=upsample,
            dropout_prob=dropout_prob,
            is3d=False,
        )


def get_model(model_config):
    model_class = get_class(model_config["name"], modules=["pytorch3dunet.unet3d.model"])
    return model_class(**model_config)


def is_model_2d(model):
    if isinstance(model, nn.DataParallel):
        model = model.module
    return isinstance(model, UNet2D)



import argparse
import os
import platform
import shutil
from enum import Enum

import torch
import yaml

from pytorch3dunet.unet3d import utils

logger = utils.get_logger("ConfigLoader")


class TorchDevice(str, Enum):
    CUDA = "cuda"
    MPS = "mps"
    CPU = "cpu"

    @classmethod
    def values(cls):
        yield from (c.value for c in cls)


def default_device() -> TorchDevice:
    logger.info("No device specified in config - determining best device automatically")
    device = TorchDevice.CPU
    if torch.cuda.is_available():
        device = TorchDevice.CUDA
    elif torch.mps.is_available():
        device = TorchDevice.MPS

    logger.info(f"Using device: {device}")
    return device


def os_dependent_dataloader_kwargs() -> dict:
    kwargs = {}
    if platform.system() == "Darwin":
        # Considerable performance improvement avoiding spawn for dataloaders and persisting loaders on MacOSX
        kwargs = {"multiprocessing_context": "forkserver", "persistent_workers": True}

    return kwargs


def override_config(args, config):
    """Overrides config params with the ones given in command line.

    Args:
        args: Command line arguments.
        config: Configuration dictionary to override.
    """

    args_dict = vars(args)
    # remove the first argument which is the config file path
    args_dict.pop("config")

    for key, value in args_dict.items():
        if value is None:
            continue
        c = config
        for k in key.split("."):
            if k not in c:
                raise ValueError(f"Invalid config key: {key}")
            if isinstance(c[k], dict):
                c = c[k]
            else:
                c[k] = value


def load_config() -> tuple[dict, str]:
    parser = argparse.ArgumentParser(description="UNet3D")
    parser.add_argument("--config", type=str, help="Path to the YAML config file", required=True)
    # add additional command line arguments for the prediction that override the ones in the config file
    parser.add_argument("--model_path", type=str, required=False)
    parser.add_argument("--loaders.output_dir", type=str, required=False)
    parser.add_argument("--loaders.test.file_paths", type=str, nargs="+", required=False)
    parser.add_argument("--loaders.test.slice_builder.patch_shape", type=int, nargs="+", required=False)
    parser.add_argument("--loaders.test.slice_builder.stride_shape", type=int, nargs="+", required=False)

    args = parser.parse_args()
    config_path = args.config
    config = _load_config_yaml(config_path)
    override_config(args, config)

    config_device = config.get("device", None)

    try:
        config["device"] = TorchDevice(config_device) if config_device is not None else default_device()
    except ValueError as e:
        raise ValueError(
            f"Config key device: {config_device} not understood -- supported values: {', '.join(TorchDevice.values())}"
        ) from e

    if config["device"] == TorchDevice.CPU:
        logger.warning("CPU mode will likely result in slow training/prediction")

    return config, config_path


def copy_config(config: dict, config_path: str):
    """Copies the config file to the checkpoint folder."""

    def _get_last_subfolder_path(path):
        subfolders = [f.path for f in os.scandir(path) if f.is_dir()]
        return max(subfolders, default=None)

    checkpoint_dir = os.path.join(config["trainer"].pop("checkpoint_dir"), "logs")
    last_run_dir = _get_last_subfolder_path(checkpoint_dir)
    config_file_name = os.path.basename(config_path)

    if last_run_dir:
        shutil.copy2(config_path, os.path.join(last_run_dir, config_file_name))


def _load_config_yaml(config_file):
    with open(config_file) as f:
        return yaml.safe_load(f)



import collections
import os

import imageio
import numpy as np
import torch

from pytorch3dunet.augment import transforms
from pytorch3dunet.datasets.utils import ConfigDataset, calculate_stats
from pytorch3dunet.unet3d.utils import get_logger

logger = get_logger("DSB2018Dataset")


def dsb_prediction_collate(batch):
    """Forms a mini-batch of (images, paths) during test time for the DSB-like datasets.

    Args:
        batch: List of samples from the dataset.

    Returns:
        Collated batch.
    """
    error_msg = "batch must contain tensors or str; found {}"
    if isinstance(batch[0], torch.Tensor):
        return torch.stack(batch, 0)
    elif isinstance(batch[0], str):
        return list(batch)
    elif isinstance(batch[0], collections.Sequence):
        # transpose tuples, i.e. [[1, 2], ['a', 'b']] to be [[1, 'a'], [2, 'b']]
        transposed = zip(*batch, strict=False)
        return [dsb_prediction_collate(samples) for samples in transposed]

    raise TypeError(error_msg.format(type(batch[0])))


class DSB2018Dataset(ConfigDataset):
    def __init__(self, root_dir, phase, transformer_config, expand_dims=True):
        assert os.path.isdir(root_dir), f"{root_dir} is not a directory"
        assert phase in ["train", "val", "test"]

        self.phase = phase

        # load raw images
        images_dir = os.path.join(root_dir, "images")
        assert os.path.isdir(images_dir)
        self.images, self.paths = self._load_files(images_dir, expand_dims)
        self.file_path = images_dir

        stats = calculate_stats(self.images, True)

        transformer = transforms.Transformer(transformer_config, stats)

        # load raw images transformer
        self.raw_transform = transformer.raw_transform()

        if phase != "test":
            # load labeled images
            masks_dir = os.path.join(root_dir, "masks")
            assert os.path.isdir(masks_dir)
            self.masks, _ = self._load_files(masks_dir, expand_dims)
            assert len(self.images) == len(self.masks)
            # load label images transformer
            self.masks_transform = transformer.label_transform()
        else:
            self.masks = None
            self.masks_transform = None

    def __getitem__(self, idx):
        if idx >= len(self):
            raise StopIteration

        img = self.images[idx]
        if self.phase != "test":
            mask = self.masks[idx]
            return self.raw_transform(img), self.masks_transform(mask)
        else:
            return self.raw_transform(img), self.paths[idx]

    def __len__(self):
        return len(self.images)

    @classmethod
    def prediction_collate(cls, batch):
        return dsb_prediction_collate(batch)

    @classmethod
    def create_datasets(cls, dataset_config, phase):
        phase_config = dataset_config[phase]
        # load data augmentation configuration
        transformer_config = phase_config["transformer"]
        # load files to process
        file_paths = phase_config["file_paths"]
        expand_dims = dataset_config.get("expand_dims", True)
        return [cls(file_paths[0], phase, transformer_config, expand_dims)]

    @staticmethod
    def _load_files(dir, expand_dims):
        files_data = []
        paths = []
        for file in os.listdir(dir):
            path = os.path.join(dir, file)
            img = np.asarray(imageio.imread(path))
            if expand_dims:
                dims = img.ndim
                img = np.expand_dims(img, axis=0)
                if dims == 3:
                    img = np.transpose(img, (3, 0, 1, 2))

            files_data.append(img)
            paths.append(path)

        return files_data, paths



import argparse

from pytorch3dunet.unet3d.config import override_config


class TestConfig:
    def test_override_config(self, test_config):
        args = argparse.Namespace(
            config=None,
            model_path="new_model.pth",
            **{
                "loaders.output_dir": "new_output_dir",
                "loaders.test.file_paths": ["file1.h5", "file2.h5"],
                "loaders.test.slice_builder.patch_shape": [64, 64, 64],
                "loaders.test.slice_builder.stride_shape": [32, 32, 32],
            },
        )

        override_config(args, test_config)

        assert test_config["model_path"] == "new_model.pth"
        assert test_config["loaders"]["output_dir"] == "new_output_dir"
        assert test_config["loaders"]["test"]["file_paths"] == ["file1.h5", "file2.h5"]
        assert test_config["loaders"]["test"]["slice_builder"]["patch_shape"] == [64, 64, 64]
        assert test_config["loaders"]["test"]["slice_builder"]["stride_shape"] == [32, 32, 32]



import h5py
import napari

pred_file1 = "/Users/adrianwolny/workspace/pytorch-3dunet/pytorch3dunet/baseline_ce/train_blob20_predictions.h5"
pred_file3 = "/Users/adrianwolny/workspace/pytorch-3dunet/pytorch3dunet/labels_weight_augs_scale_ce/train_blob20_predictions.h5"
pred_file4 = "/Users/adrianwolny/workspace/pytorch-3dunet/pytorch3dunet/merged_labels_weight_augs_ce/train_blob20_predictions.h5"
raw_file = "/Users/adrianwolny/workspace/pytorch-3dunet/pytorch3dunet/baseline_ce/train_blob20.h5"

with h5py.File(raw_file, "r") as f:
    raw = f["raw"][:]
    label = f["label"][:]

with h5py.File(pred_file1, "r") as f:
    baseline_ce = f["predictions"][:]

with h5py.File(pred_file3, "r") as f:
    merged_labels_weight_ce = f["predictions"][:]

with h5py.File(pred_file4, "r") as f:
    merged_labels_weight_augs_ce = f["predictions"][:]

viewer = napari.Viewer()
viewer.add_image(raw, name="raw")
viewer.add_labels(label, name="label")
viewer.add_labels(baseline_ce, name="baseline_ce")
viewer.add_labels(merged_labels_weight_ce, name="labels_weight_augs_scale_ce")
viewer.add_labels(merged_labels_weight_augs_ce, name="merged_labels_weight_augs_ce")

napari.run()



import os
from pathlib import Path
from tempfile import NamedTemporaryFile

import h5py
import numpy as np

from pytorch3dunet.datasets.utils import get_test_loaders
from pytorch3dunet.predict import get_predictor
from pytorch3dunet.unet3d.model import get_model


def _run_prediction(test_config, tmpdir, shape):
    # Add output dir
    test_config["loaders"]["output_dir"] = tmpdir
    # create random dataset
    tmp = NamedTemporaryFile(delete=False)
    with h5py.File(tmp.name, "w") as f:
        f.create_dataset("raw", data=np.random.rand(*shape))
        f.create_dataset("label", data=np.random.randint(0, 2, shape).astype("uint8"))
    # Add input file
    test_config["loaders"]["test"]["file_paths"] = [tmp.name]
    # Create the model with random weights
    model = get_model(test_config["model"])

    model = model.to(test_config["device"])
    results = []
    for test_loader in get_test_loaders(test_config):
        predictor = get_predictor(model, test_config)
        # run the model prediction on the entire dataset and save to the 'output_file' H5
        result = predictor(test_loader)
        if result is not None:
            results.append(result)
    if results:
        return tmp, results
    return tmp


def _get_result_shape(result_path: str, dataset_name: str = "predictions"):
    with h5py.File(result_path, "r") as f:
        return f[dataset_name].shape


class TestPredictor:
    def test_3d_predictor(self, tmpdir, test_config):
        tmp = _run_prediction(test_config, tmpdir, shape=(32, 64, 64))
        output_filename = os.path.split(tmp.name)[1] + "_predictions.h5"
        output_path = Path(tmpdir) / output_filename
        assert output_path.exists()
        assert _get_result_shape(output_path) == (2, 32, 64, 64)

    def test_2d_predictor(self, tmpdir, test_config_2d):
        tmp = _run_prediction(test_config_2d, tmpdir, shape=(3, 1, 256, 256))
        output_filename = os.path.split(tmp.name)[1] + "_predictions.h5"
        output_path = Path(tmpdir) / output_filename
        assert output_path.exists()
        assert _get_result_shape(output_path) == (2, 1, 256, 256)

    def test_predictor_save_segmentation(self, tmpdir, test_config):
        test_config["predictor"]["save_segmentation"] = True
        tmp = _run_prediction(test_config, tmpdir, shape=(32, 64, 64))
        output_filename = os.path.split(tmp.name)[1] + "_predictions.h5"
        output_path = Path(tmpdir) / output_filename
        assert output_path.exists()
        # check that the output segmentation is saved, with the channel dimension reduced via argmax operation
        assert _get_result_shape(output_path) == (32, 64, 64)

    def test_performance_metric(self, tmpdir, test_config):
        test_config["predictor"]["save_segmentation"] = True
        test_config["predictor"]["performance_metric"] = "mean_iou"
        test_config["predictor"]["gt_internal_path"] = "label"
        tmp, results = _run_prediction(test_config, tmpdir, shape=(32, 64, 64))

        output_filename = os.path.split(tmp.name)[1] + "_predictions.h5"
        output_path = Path(tmpdir) / output_filename
        assert output_path.exists()
        # check that the output segmentation is saved, with the channel dimension reduced via argmax operation
        assert _get_result_shape(output_path) == (32, 64, 64)
        # assert results array is non-zero
        assert np.mean(results) > 0

    def test_lazy_predictor(self, tmpdir, test_config):
        test_config["predictor"]["name"] = "LazyPredictor"
        tmp = _run_prediction(test_config, tmpdir, shape=(32, 64, 64))
        output_filename = os.path.split(tmp.name)[1] + "_predictions.h5"
        output_path = Path(tmpdir) / output_filename
        assert output_path.exists()
        # check that the output segmentation is saved, with the channel dimension reduced via argmax operation
        assert _get_result_shape(output_path) == (2, 32, 64, 64)
        # assert results array is non-zero
        with h5py.File(output_path, "r") as f:
            assert np.count_nonzero(f["predictions"][...]) > 0

    def test_performance_metric_lazy_predictor(self, tmpdir, test_config):
        test_config["predictor"]["name"] = "LazyPredictor"
        test_config["predictor"]["save_segmentation"] = True
        test_config["predictor"]["performance_metric"] = "mean_iou"
        test_config["predictor"]["gt_internal_path"] = "label"
        tmp, results = _run_prediction(test_config, tmpdir, shape=(32, 64, 64))

        output_filename = os.path.split(tmp.name)[1] + "_predictions.h5"
        output_path = Path(tmpdir) / output_filename
        assert output_path.exists()
        # check that the output segmentation is saved, with the channel dimension reduced via argmax operation
        assert _get_result_shape(output_path) == (32, 64, 64)
        # assert results array is non-zero
        assert np.mean(results) > 0



from abc import abstractmethod
from collections.abc import Iterable
from itertools import chain
from pathlib import Path

import h5py
import numpy as np

import pytorch3dunet.augment.transforms as transforms
from pytorch3dunet.datasets.utils import ConfigDataset, RandomScaler, calculate_stats, get_slice_builder, mirror_pad
from pytorch3dunet.unet3d.utils import get_logger

logger = get_logger("HDF5Dataset")


def _create_padded_indexes(indexes: tuple, halo_shape: tuple):
    """Create padded indexes by extending each slice in `indexes` by the corresponding `halo_shape`."""
    if sum(halo_shape) == 0:
        return indexes
    return tuple(slice(index.start, index.stop + 2 * halo) for index, halo in zip(indexes, halo_shape, strict=True))


def traverse_h5_paths(file_paths: list[str]) -> list[str]:
    """Traverse the given list of file paths and directories to find all H5 files."""
    assert isinstance(file_paths, list)
    results = []
    for file_path in file_paths:
        file_path = Path(file_path)
        if file_path.is_dir():
            # if file_path is a directory take all H5 files in that directory
            iters = [file_path.glob(ext) for ext in ["*.h5", "*.hdf", "*.hdf5", "*.hd5"]]
            for fp in chain(*iters):
                results.append(str(fp))
        else:
            results.append(str(file_path))
    return results


class AbstractHDF5Dataset(ConfigDataset):
    """
    Implementation of torch.utils.data.Dataset backed by the HDF5 files, which iterates over the raw and label datasets
    patch by patch with a given stride. It's an abstract class for the standard and lazy implementations.

    Args:
        file_path (str): path to H5 file containing raw data and (optional) ground truth labels
        phase (str): 'train' for training, 'val' for validation, 'test' for testing
        slice_builder_config (dict): configuration of the SliceBuilder
        transformer_config (dict): data augmentation configuration
        raw_internal_path (str): H5 internal path to the raw dataset
        label_internal_path (str): H5 internal path to the label dataset
        global_normalization (bool): if True, the mean and std of the raw data will be calculated over the whole dataset
        random_scale (int): if not None, the raw data will be randomly shifted by a value in the range
            [-random_scale, random_scale] in each dimension and then scaled to the original patch shape
        random_scale_probability (float): probability of executing the random scale on a patch
    """

    def __init__(
        self,
        file_path: str,
        phase: str,
        slice_builder_config: dict,
        transformer_config: dict,
        raw_internal_path: str = "raw",
        label_internal_path: str = "label",
        global_normalization: bool = False,
        random_scale: int | None = None,
        random_scale_probability: float = 0.5,
    ):
        assert phase in ["train", "val", "test"]
        logger.info(f"Creating {self.__class__.__name__} for {phase} phase from {file_path}")
        self.phase = phase
        self.file_path = file_path
        self.raw_internal_path = raw_internal_path
        self.label_internal_path = label_internal_path

        self.halo_shape = tuple(slice_builder_config.get("halo_shape", [0, 0, 0]))

        if global_normalization:
            logger.info("Calculating mean and std of the raw data...")
            with h5py.File(file_path, "r") as f:
                raw = f[raw_internal_path][:]
                stats = calculate_stats(raw)
        else:
            stats = calculate_stats(None, True)

        self.transformer = transforms.Transformer(transformer_config, stats)
        self.raw_transform = self.transformer.raw_transform()

        if phase != "test":
            # create label transform only in train/val phase
            self.label_transform = self.transformer.label_transform()
        else:
            # 'test' phase used only for predictions so ignore the label dataset
            self.label = None

            if self.halo_shape == (0, 0, 0):
                logger.warning(
                    "Found halo shape to be (0, 0, 0). This might lead to checkerboard artifacts in the "
                    "prediction. Consider using a non-zero halo shape, e.g. 'halo_shape: [8, 8, 8]' in "
                    "the slice_builder configuration."
                )

        with h5py.File(file_path, "r") as f:
            raw = f[raw_internal_path]
            if raw.ndim == 3:
                self.volume_shape = raw.shape
            else:
                self.volume_shape = raw.shape[1:]
            label = f[label_internal_path] if phase != "test" else None
            # check that raw and label shapes match
            if label is not None:
                if label.ndim == 3:
                    assert label.shape == self.volume_shape, "Raw and label shapes do not match"
                else:
                    assert label.shape[1:] == self.volume_shape, "Raw and label shapes do not match"

            logger.info(f"Volume shape: {self.volume_shape}. Creating slices...")
            # build slice indices for raw and label data sets
            slice_builder_config["lazy_loader"] = self.is_lazy()  # pass this information to the slice builder
            slice_builder = get_slice_builder(raw, label, slice_builder_config)
            self.raw_slices = slice_builder.raw_slices
            self.label_slices = slice_builder.label_slices

        if random_scale is not None:
            assert isinstance(random_scale, int), "random_scale must be an integer"
            stride_shape = slice_builder_config.get("stride_shape")
            assert all(random_scale < stride for stride in stride_shape), (
                f"random_scale {random_scale} must be smaller than each of the strides {stride_shape}"
            )
            patch_shape = slice_builder_config.get("patch_shape")
            self.random_scaler = RandomScaler(random_scale, patch_shape, self.volume_shape, random_scale_probability)
            logger.info(f"Using RandomScaler with offset range {random_scale}")
        else:
            self.random_scaler = None

        self.patch_count = len(self.raw_slices)

    @abstractmethod
    def get_raw_patch(self, idx: int) -> np.ndarray:
        raise NotImplementedError

    @abstractmethod
    def get_label_patch(self, idx) -> np.ndarray:
        raise NotImplementedError

    @abstractmethod
    def get_raw_padded_patch(self, idx) -> np.ndarray:
        raise NotImplementedError

    @abstractmethod
    def is_lazy(self) -> bool:
        raise NotImplementedError

    def __getitem__(self, idx: int) -> tuple:
        if idx >= len(self):
            raise StopIteration

        raw_idx = self.raw_slices[idx]

        if self.phase == "test":
            if len(raw_idx) == 4:
                # discard the channel dimension in the slices: predictor requires only the spatial dimensions of the volume
                raw_idx = raw_idx[1:]  # Remove the first element if raw_idx has 4 elements
                raw_idx_padded = (slice(None),) + _create_padded_indexes(raw_idx, self.halo_shape)
            else:
                raw_idx_padded = _create_padded_indexes(raw_idx, self.halo_shape)

            # use halo padding in the 'test' phase to avoid edge artifacts
            padded_patch = self.get_raw_padded_patch(raw_idx_padded)
            raw_patch_transformed = self.raw_transform(padded_patch)
            # return padded patch, and the original (non-padded) index for placing the prediction back into the volume
            # predictor is responsible for removing the halo from the prediction
            return raw_patch_transformed, raw_idx
        else:
            label_idx = self.label_slices[idx]

            if self.random_scaler is not None:
                # randomize the indices
                raw_idx, label_idx = self.random_scaler.randomize_indices(raw_idx, label_idx)

            raw_patch_transformed = self.raw_transform(self.get_raw_patch(raw_idx))
            label_patch_transformed = self.label_transform(self.get_label_patch(label_idx))

            if self.random_scaler is not None:
                # scale patches back to the original patch size
                raw_patch_transformed, label_patch_transformed = self.random_scaler.rescale_patches(
                    raw_patch_transformed, label_patch_transformed
                )
            # return the transformed raw and label patches
            return raw_patch_transformed, label_patch_transformed

    def __len__(self) -> int:
        return self.patch_count

    @classmethod
    def create_datasets(cls, dataset_config: dict, phase: str) -> Iterable["AbstractHDF5Dataset"]:
        phase_config = dataset_config[phase]

        # load data augmentation configuration
        transformer_config = phase_config["transformer"]
        # load slice builder config
        slice_builder_config = phase_config["slice_builder"]
        # load files to process
        file_paths = phase_config["file_paths"]
        # file_paths may contain both files and directories; if the file_path is a directory all H5 files inside
        # are going to be included in the final file_paths
        file_paths = traverse_h5_paths(file_paths)

        # create dataset for each file path
        for file_path in file_paths:
            yield cls(
                file_path=file_path,
                phase=phase,
                slice_builder_config=slice_builder_config,
                transformer_config=transformer_config,
                raw_internal_path=dataset_config.get("raw_internal_path", "raw"),
                label_internal_path=dataset_config.get("label_internal_path", "label"),
                global_normalization=dataset_config.get("global_normalization", False),
                random_scale=dataset_config.get("random_scale", None),
                random_scale_probability=dataset_config.get("random_scale_probability", 0.5),
            )


class StandardHDF5Dataset(AbstractHDF5Dataset):
    """Implementation of the HDF5 dataset which loads the data from the H5 files into the memory.
    Fast but might consume a lot of memory.
    """

    def __init__(
        self,
        file_path: str,
        phase: str,
        slice_builder_config: dict,
        transformer_config: dict,
        raw_internal_path: str = "raw",
        label_internal_path: str = "label",
        global_normalization: bool = False,
        random_scale: int = None,
        random_scale_probability: float = 0.5,
    ):
        super().__init__(
            file_path=file_path,
            phase=phase,
            slice_builder_config=slice_builder_config,
            transformer_config=transformer_config,
            raw_internal_path=raw_internal_path,
            label_internal_path=label_internal_path,
            global_normalization=global_normalization,
            random_scale=random_scale,
            random_scale_probability=random_scale_probability,
        )
        self._raw = None
        self._raw_padded = None
        self._label = None

    def get_raw_patch(self, idx):
        # load lazily on the first request
        if self._raw is None:
            with h5py.File(self.file_path, "r") as f:
                assert self.raw_internal_path in f, f"Dataset {self.raw_internal_path} not found in {self.file_path}"
                self._raw = f[self.raw_internal_path][:]
        return self._raw[idx]

    def get_label_patch(self, idx):
        if self._label is None:
            with h5py.File(self.file_path, "r") as f:
                assert self.label_internal_path in f, (
                    f"Dataset {self.label_internal_path} not found in {self.file_path}"
                )
                self._label = f[self.label_internal_path][:]
        return self._label[idx]

    def get_raw_padded_patch(self, idx):
        if self._raw_padded is None:
            with h5py.File(self.file_path, "r") as f:
                assert self.raw_internal_path in f, f"Dataset {self.raw_internal_path} not found in {self.file_path}"
                self._raw_padded = mirror_pad(f[self.raw_internal_path][:], self.halo_shape)
        return self._raw_padded[idx]

    def is_lazy(self) -> bool:
        return False


class LazyHDF5Dataset(AbstractHDF5Dataset):
    """Implementation of the HDF5 dataset which loads the data lazily.
    It's slower, but has a low memory footprint.
    """

    def __init__(
        self,
        file_path: str,
        phase: str,
        slice_builder_config: dict,
        transformer_config: dict,
        raw_internal_path: str = "raw",
        label_internal_path: str = "label",
        global_normalization: bool = False,
        random_scale: int = None,
        random_scale_probability: float = 0.5,
    ):
        super().__init__(
            file_path=file_path,
            phase=phase,
            slice_builder_config=slice_builder_config,
            transformer_config=transformer_config,
            raw_internal_path=raw_internal_path,
            label_internal_path=label_internal_path,
            global_normalization=global_normalization,
            random_scale=random_scale,
            random_scale_probability=random_scale_probability,
        )

        logger.info("Using LazyHDF5Dataset")

    def get_raw_patch(self, idx: int) -> np.ndarray:
        with h5py.File(self.file_path, "r") as f:
            return f[self.raw_internal_path][idx]

    def get_label_patch(self, idx: int) -> np.ndarray:
        with h5py.File(self.file_path, "r") as f:
            return f[self.label_internal_path][idx]

    def get_raw_padded_patch(self, idx: int) -> np.ndarray:
        with h5py.File(self.file_path, "r+") as f:
            if "raw_padded" in f:
                return f["raw_padded"][idx]

            logger.info(f"Creating 'raw_padded' dataset in {self.file_path}")
            raw = f[self.raw_internal_path][:]
            raw_padded = mirror_pad(raw, self.halo_shape)
            f.create_dataset("raw_padded", data=raw_padded, compression="gzip")
            return raw_padded[idx]

    def is_lazy(self) -> bool:
        return True



import numpy as np
from skimage.metrics import contingency_table


def precision(tp, fp, fn):
    """Computes precision from true positives, false positives, and false negatives.

    Args:
        tp: Number of true positives.
        fp: Number of false positives.
        fn: Number of false negatives.

    Returns:
        Precision score.
    """
    return tp / (tp + fp) if tp > 0 else 0


def recall(tp, fp, fn):
    """Computes recall from true positives, false positives, and false negatives.

    Args:
        tp: Number of true positives.
        fp: Number of false positives.
        fn: Number of false negatives.

    Returns:
        Recall score.
    """
    return tp / (tp + fn) if tp > 0 else 0


def accuracy(tp, fp, fn):
    """Computes accuracy from true positives, false positives, and false negatives.

    Args:
        tp: Number of true positives.
        fp: Number of false positives.
        fn: Number of false negatives.

    Returns:
        Accuracy score.
    """
    return tp / (tp + fp + fn) if tp > 0 else 0


def f1(tp, fp, fn):
    """Computes F1 score from true positives, false positives, and false negatives.

    Args:
        tp: Number of true positives.
        fp: Number of false positives.
        fn: Number of false negatives.

    Returns:
        F1 score.
    """
    return (2 * tp) / (2 * tp + fp + fn) if tp > 0 else 0


def _relabel(input):
    _, unique_labels = np.unique(input, return_inverse=True)
    return unique_labels.reshape(input.shape)


def _iou_matrix(gt, seg):
    # relabel gt and seg for smaller memory footprint of contingency table
    gt = _relabel(gt)
    seg = _relabel(seg)

    # get number of overlapping pixels between GT and SEG
    n_inter = contingency_table(gt, seg).toarray()

    # number of pixels for GT instances
    n_gt = n_inter.sum(axis=1, keepdims=True)
    # number of pixels for SEG instances
    n_seg = n_inter.sum(axis=0, keepdims=True)

    # number of pixels in the union between GT and SEG instances
    n_union = n_gt + n_seg - n_inter

    iou_matrix = n_inter / n_union
    # make sure that the values are within [0,1] range
    assert 0 <= np.min(iou_matrix) <= np.max(iou_matrix) <= 1

    return iou_matrix


class SegmentationMetrics:
    """
    Computes precision, recall, accuracy, f1 score for a given ground truth and predicted segmentation.
    Contingency table for a given ground truth and predicted segmentation is computed eagerly upon construction
    of the instance of `SegmentationMetrics`.

    Args:
        gt (ndarray): ground truth segmentation
        seg (ndarray): predicted segmentation
    """

    def __init__(self, gt, seg):
        self.iou_matrix = _iou_matrix(gt, seg)

    def metrics(self, iou_threshold):
        """Computes precision, recall, accuracy, f1 score at a given IoU threshold.

        Args:
            iou_threshold: IoU threshold for detection.

        Returns:
            Dictionary containing precision, recall, accuracy, and f1 scores.
        """
        # ignore background
        iou_matrix = self.iou_matrix[1:, 1:]
        detection_matrix = (iou_matrix > iou_threshold).astype(np.uint8)
        n_gt, n_seg = detection_matrix.shape

        # if the iou_matrix is empty or all values are 0
        trivial = min(n_gt, n_seg) == 0 or np.all(detection_matrix == 0)
        if trivial:
            tp = fp = fn = 0
        else:
            # count non-zero rows to get the number of TP
            tp = np.count_nonzero(detection_matrix.sum(axis=1))
            # count zero rows to get the number of FN
            fn = n_gt - tp
            # count zero columns to get the number of FP
            fp = n_seg - np.count_nonzero(detection_matrix.sum(axis=0))

        return {
            "precision": precision(tp, fp, fn),
            "recall": recall(tp, fp, fn),
            "accuracy": accuracy(tp, fp, fn),
            "f1": f1(tp, fp, fn),
        }


class Accuracy:
    """Computes accuracy between ground truth and predicted segmentation at a given threshold value.

    Defined as: AC = TP / (TP + FP + FN).
    Kaggle DSB2018 calls it Precision, see:
    https://www.kaggle.com/stkbailey/step-by-step-explanation-of-scoring-metric.

    Args:
        iou_threshold: IoU threshold for detection.
    """

    def __init__(self, iou_threshold):
        self.iou_threshold = iou_threshold

    def __call__(self, input_seg, gt_seg):
        metrics = SegmentationMetrics(gt_seg, input_seg).metrics(self.iou_threshold)
        return metrics["accuracy"]


class AveragePrecision:
    """Average precision taken for the IoU range (0.5, 0.95) with a step of 0.05.

    As defined in: https://www.kaggle.com/stkbailey/step-by-step-explanation-of-scoring-metric
    """

    def __init__(self):
        self.iou_range = np.linspace(0.50, 0.95, 10)

    def __call__(self, input_seg, gt_seg):
        # compute contingency_table
        sm = SegmentationMetrics(gt_seg, input_seg)
        # compute accuracy for each threshold
        acc = [sm.metrics(iou)["accuracy"] for iou in self.iou_range]
        # return the average
        return np.mean(acc)



import importlib
import random
from collections.abc import Callable

import numpy as np
import torch
from scipy.ndimage import convolve, gaussian_filter, map_coordinates, rotate
from skimage import exposure, measure
from skimage.filters import gaussian
from skimage.segmentation import find_boundaries

# WARN: use fixed random state for reproducibility; if you want to randomize on each run seed with `time.time()` e.g.
GLOBAL_RANDOM_STATE = np.random.RandomState(47)


class Compose:
    """Composes several transforms together."""

    def __init__(self, transforms: list[Callable[[np.ndarray], np.ndarray]]):
        self.transforms = transforms

    def __call__(self, m):
        for t in self.transforms:
            m = t(m)
        return m


class RandomFlip:
    """Randomly flips the image across the given axes.

    Image can be either 3D (DxHxW) or 4D (CxDxHxW).
    When creating make sure that the provided RandomStates are consistent between raw and labeled datasets,
    otherwise the models won't converge.

    Args:
        random_state: Random state for reproducibility.
        axis_prob: Probability of flipping along each axis. Default: 0.5.
    """

    def __init__(self, random_state: np.random.RandomState, axis_prob: float = 0.5, **kwargs):
        assert random_state is not None, "RandomState cannot be None"
        self.random_state = random_state
        self.axes = (0, 1, 2)
        self.axis_prob = axis_prob

    def __call__(self, m: np.ndarray) -> np.ndarray:
        assert m.ndim in [3, 4], "Supports only 3D (DxHxW) or 4D (CxDxHxW) images"

        for axis in self.axes:
            if self.random_state.uniform() > self.axis_prob:
                if m.ndim == 3:
                    m = np.flip(m, axis)
                else:
                    channels = [np.flip(m[c], axis) for c in range(m.shape[0])]
                    m = np.stack(channels, axis=0)

        return m


class RandomRotate90:
    """Rotate an array by 90 degrees around a randomly chosen plane.

    Image can be either 3D (DxHxW) or 4D (CxDxHxW).
    When creating make sure that the provided RandomStates are consistent between raw and labeled datasets,
    otherwise the models won't converge.

    IMPORTANT: Assumes DHW axis order (that's why rotation is performed across (1,2) axis).

    Args:
        random_state: Random state for reproducibility.
    """

    def __init__(self, random_state: np.random.RandomState, **kwargs):
        self.random_state = random_state
        # always rotate around z-axis
        self.axis = (1, 2)

    def __call__(self, m: np.ndarray) -> np.ndarray:
        assert m.ndim in [3, 4], "Supports only 3D (DxHxW) or 4D (CxDxHxW) images"

        # pick number of rotations at random
        k = self.random_state.randint(0, 4)
        # rotate k times around a given plane
        if m.ndim == 3:
            m = np.rot90(m, k, self.axis)
        else:
            channels = [np.rot90(m[c], k, self.axis) for c in range(m.shape[0])]
            m = np.stack(channels, axis=0)

        return m


class RandomRotate:
    """Rotate an array by a random degrees from (-angle_spectrum, angle_spectrum) interval.

    Rotation axis is picked at random from the list of provided axes.

    Args:
        random_state: Random state for reproducibility.
        angle_spectrum: Maximum rotation angle. Default: 30.
        axes: List of rotation axes. Default: [(1, 0), (2, 1), (2, 0)].
        mode: Interpolation mode. Default: 'reflect'.
        order: Interpolation order. Default: 0.
    """

    def __init__(
        self,
        random_state: np.random.RandomState,
        angle_spectrum: int = 30,
        axes: list = None,
        mode: str = "reflect",
        order: int = 0,
        **kwargs,
    ):
        if axes is None:
            axes = [(1, 0), (2, 1), (2, 0)]
        else:
            assert isinstance(axes, list) and len(axes) > 0

        self.random_state = random_state
        self.angle_spectrum = angle_spectrum
        self.axes = axes
        self.mode = mode
        self.order = order

    def __call__(self, m: np.ndarray) -> np.ndarray:
        axis = self.axes[self.random_state.randint(len(self.axes))]
        angle = self.random_state.randint(-self.angle_spectrum, self.angle_spectrum)

        if m.ndim == 3:
            m = rotate(m, angle, axes=axis, reshape=False, order=self.order, mode=self.mode, cval=-1)
        else:
            channels = [
                rotate(m[c], angle, axes=axis, reshape=False, order=self.order, mode=self.mode, cval=-1)
                for c in range(m.shape[0])
            ]
            m = np.stack(channels, axis=0)

        return m


class RandomContrast:
    """Adjust contrast by scaling each voxel to `mean + alpha * (v - mean)`.

    Args:
        random_state: Random state for reproducibility.
        alpha: Range of contrast adjustment factor. Default: (0.5, 1.5).
        mean: Mean value for contrast adjustment. Default: 0.0.
        execution_probability: Probability of applying this transform. Default: 0.1.
    """

    def __init__(
        self,
        random_state: np.random.RandomState,
        alpha: tuple[float, float] = (0.5, 1.5),
        mean: float = 0.0,
        execution_probability: float = 0.1,
        **kwargs,
    ):
        self.random_state = random_state
        assert len(alpha) == 2
        self.alpha = alpha
        self.mean = mean
        self.execution_probability = execution_probability

    def __call__(self, m: np.ndarray) -> np.ndarray:
        if self.random_state.uniform() < self.execution_probability:
            alpha = self.random_state.uniform(self.alpha[0], self.alpha[1])
            result = self.mean + alpha * (m - self.mean)
            return np.clip(result, -1, 1)

        return m


class RandomGammaCorrection:
    """Adjust contrast by scaling each voxel to `v ** gamma`.

    Args:
        random_state: Random state for reproducibility.
        gamma: Range of gamma values. Default: (0.5, 1.5).
        execution_probability: Probability of applying this transform. Default: 0.1.
    """

    def __init__(
        self,
        random_state: np.random.RandomState,
        gamma: tuple[float, float] = (0.5, 1.5),
        execution_probability: float = 0.1,
        **kwargs,
    ):
        self.random_state = random_state
        assert len(gamma) == 2
        self.gamma = gamma
        self.execution_probability = execution_probability

    def __call__(self, m: np.ndarray) -> np.ndarray:
        if self.random_state.uniform() < self.execution_probability:
            # rescale intensity values to [0, 1]
            m = exposure.rescale_intensity(m, out_range=(0, 1))
            gamma = self.random_state.uniform(self.gamma[0], self.gamma[1])
            return exposure.adjust_gamma(m, gamma)

        return m


class ElasticDeformation:
    """Apply elastic deformations of 3D patches on a per-voxel mesh.
    This augmentation  relatively slow (~1s per patch of size 64x200x200), so use multiple workers in the DataLoader.
    Remember to use spline_order=0 when transforming the labels.

    Assumes ZYX axis order (or CZYX if the data is 4D).
    Based on: https://github.com/fcalvet/image_tools/blob/master/image_augmentation.py#L62

    Args:
        random_state: Random state for reproducibility.
        spline_order: The order of spline interpolation (use 0 for labeled images).
        alpha: Scaling factor for deformations. Default: 2000.
        sigma: Smoothing factor for Gaussian filter. Default: 50.
        execution_probability: Probability of executing this transform. Default: 0.1.
        apply_3d: If True apply deformations in each axis. Default: True.
    """

    def __init__(
        self,
        random_state: np.random.RandomState,
        spline_order: int,
        alpha: int = 2000,
        sigma: int = 50,
        execution_probability: float = 0.1,
        apply_3d: bool = True,
        **kwargs,
    ):
        self.random_state = random_state
        self.spline_order = spline_order
        self.alpha = alpha
        self.sigma = sigma
        self.execution_probability = execution_probability
        self.apply_3d = apply_3d

    def __call__(self, m: np.ndarray) -> np.ndarray:
        if self.random_state.uniform() < self.execution_probability:
            assert m.ndim in [3, 4]

            if m.ndim == 3:
                volume_shape = m.shape
            else:
                volume_shape = m[0].shape

            if self.apply_3d:
                dz = gaussian_filter(self.random_state.randn(*volume_shape), self.sigma, mode="reflect") * self.alpha
            else:
                dz = np.zeros_like(m)

            dy, dx = [
                gaussian_filter(self.random_state.randn(*volume_shape), self.sigma, mode="reflect") * self.alpha
                for _ in range(2)
            ]

            z_dim, y_dim, x_dim = volume_shape
            z, y, x = np.meshgrid(np.arange(z_dim), np.arange(y_dim), np.arange(x_dim), indexing="ij")
            indices = z + dz, y + dy, x + dx

            if m.ndim == 3:
                return map_coordinates(m, indices, order=self.spline_order, mode="reflect")
            else:
                channels = [map_coordinates(c, indices, order=self.spline_order, mode="reflect") for c in m]
                return np.stack(channels, axis=0)

        return m


class CropToFixed:
    """Crop or pad the input array to a fixed size.

    Args:
        random_state: Random state for reproducibility.
        size: Desired output size (y, x). Default: (256, 256).
        centered: If True, always crop/pad around the center. Default: False.
    """

    def __init__(
        self, random_state: np.random.RandomState, size: tuple[int, int] = (256, 256), centered: bool = False, **kwargs
    ):
        self.random_state = random_state
        self.crop_y, self.crop_x = size
        self.centered = centered

    def __call__(self, m: np.ndarray) -> np.ndarray:
        def _padding(pad_total: int) -> tuple[int, int]:
            half_total = pad_total // 2
            return half_total, pad_total - half_total

        def _rand_range_and_pad(crop_size: int, max_size: int) -> tuple[int, tuple[int, int]]:
            """
            Returns a tuple:
                max_value for the corner dimension. The corner dimension is chosen as `self.random_state(max_value)`
                pad: padding in both directions; if crop_size is lt max_size the pad is 0
            """
            if crop_size < max_size:
                return max_size - crop_size, (0, 0)
            else:
                return 1, _padding(crop_size - max_size)

        def _start_and_pad(crop_size: int, max_size: int) -> tuple[int, tuple[int, int]]:
            if crop_size < max_size:
                return (max_size - crop_size) // 2, (0, 0)
            else:
                return 0, _padding(crop_size - max_size)

        assert m.ndim in (3, 4)
        if m.ndim == 3:
            _, y, x = m.shape
        else:
            _, _, y, x = m.shape

        if not self.centered:
            y_range, y_pad = _rand_range_and_pad(self.crop_y, y)
            x_range, x_pad = _rand_range_and_pad(self.crop_x, x)

            y_start = self.random_state.randint(y_range)
            x_start = self.random_state.randint(x_range)

        else:
            y_start, y_pad = _start_and_pad(self.crop_y, y)
            x_start, x_pad = _start_and_pad(self.crop_x, x)

        if m.ndim == 3:
            result = m[:, y_start : y_start + self.crop_y, x_start : x_start + self.crop_x]
            return np.pad(result, pad_width=((0, 0), y_pad, x_pad), mode="reflect")
        else:
            channels = []
            for c in range(m.shape[0]):
                result = m[c][:, y_start : y_start + self.crop_y, x_start : x_start + self.crop_x]
                channels.append(np.pad(result, pad_width=((0, 0), y_pad, x_pad), mode="reflect"))
            return np.stack(channels, axis=0)


class StandardLabelToBoundary:
    """Converts a given volumetric label array to binary mask corresponding to borders between labels.

    Args:
        ignore_index: Label to ignore in the output.
        append_label: If True, stack the borders and original labels across channel dim. Default: False.
        mode: Boundary detection mode. Default: 'thick'.
        foreground: If True, include foreground mask (i.e everything greater than 0) in the first channel of the result.
            Default: False.
    """

    def __init__(
        self,
        ignore_index: int = None,
        append_label: bool = False,
        mode: str = "thick",
        foreground: bool = False,
        **kwargs,
    ):
        self.ignore_index = ignore_index
        self.append_label = append_label
        self.mode = mode
        self.foreground = foreground

    def __call__(self, m: np.ndarray) -> np.ndarray:
        assert m.ndim == 3

        boundaries = find_boundaries(m, connectivity=2, mode=self.mode)
        boundaries = boundaries.astype("int32")

        results = []
        if self.foreground:
            foreground = (m > 0).astype("uint8")
            results.append(_recover_ignore_index(foreground, m, self.ignore_index))

        results.append(_recover_ignore_index(boundaries, m, self.ignore_index))

        if self.append_label:
            # append original input data
            results.append(m)

        return np.stack(results, axis=0)


class BlobsToMask:
    """Returns binary mask from labeled image of blob like objects.

    Every label greater than 0 is treated as foreground.

    Args:
        append_label: If True, append original labels. Default: False.
        boundary: If True, compute boundaries. Default: False.
        cross_entropy: If True, use cross entropy format. Default: False.
    """

    def __init__(self, append_label: bool = False, boundary: bool = False, cross_entropy: bool = False, **kwargs):
        self.cross_entropy = cross_entropy
        self.boundary = boundary
        self.append_label = append_label

    def __call__(self, m):
        assert m.ndim == 3

        # get the segmentation mask
        mask = (m > 0).astype("uint8")
        results = [mask]

        if self.boundary:
            outer = find_boundaries(m, connectivity=2, mode="outer")
            if self.cross_entropy:
                # boundary is class 2
                mask[outer > 0] = 2
                results = [mask]
            else:
                results.append(outer)

        if self.append_label:
            results.append(m)

        return np.stack(results, axis=0)


class AbstractLabelToBoundary:
    """Abstract base class for label to boundary conversion.

    Args:
        ignore_index: Label to be ignored in the output, i.e. after computing the boundary the label
            ignore_index will be restored where it was in the patch originally.
        aggregate_affinities: Aggregate affinities with the same offset across Z,Y,X axes. Default: False.
        append_label: If True append the original ground truth labels to the last channel. Default: False.
    """

    AXES_TRANSPOSE = [
        (0, 1, 2),  # X
        (0, 2, 1),  # Y
        (2, 0, 1),  # Z
    ]

    def __init__(
        self, ignore_index: int = None, aggregate_affinities: bool = False, append_label: bool = False, **kwargs
    ):
        self.ignore_index = ignore_index
        self.aggregate_affinities = aggregate_affinities
        self.append_label = append_label

    def __call__(self, m: np.ndarray) -> np.ndarray:
        """Extract boundaries from a given 3D label tensor.

        Args:
            m: Input 3D tensor.

        Returns:
            Binary mask, with 1-label corresponding to the boundary and 0-label corresponding to the background.
        """
        assert m.ndim == 3

        kernels = self.get_kernels()
        boundary_arr = [np.where(np.abs(convolve(m, kernel)) > 0, 1, 0) for kernel in kernels]
        channels = np.stack(boundary_arr)
        results = []
        if self.aggregate_affinities:
            assert len(kernels) % 3 == 0, "Number of kernels must be divided by 3 (one kernel per offset per Z,Y,X axes"
            # aggregate affinities with the same offset
            for i in range(0, len(kernels), 3):
                # merge across X,Y,Z axes (logical OR)
                xyz_aggregated_affinities = np.logical_or.reduce(channels[i : i + 3, ...]).astype(np.int32)
                # recover ignore index
                xyz_aggregated_affinities = _recover_ignore_index(xyz_aggregated_affinities, m, self.ignore_index)
                results.append(xyz_aggregated_affinities)
        else:
            results = [_recover_ignore_index(channels[i], m, self.ignore_index) for i in range(channels.shape[0])]

        if self.append_label:
            # append original input data
            results.append(m)

        # stack across channel dim
        return np.stack(results, axis=0)

    @staticmethod
    def create_kernel(axis: int | tuple, offset: int) -> np.ndarray:
        # create conv kernel
        k_size = offset + 1
        k = np.zeros((1, 1, k_size), dtype=np.int32)
        k[0, 0, 0] = 1
        k[0, 0, offset] = -1
        return np.transpose(k, axis)

    def get_kernels(self):
        raise NotImplementedError


class RandomLabelToAffinities(AbstractLabelToBoundary):
    """Converts a given volumetric label array to binary mask corresponding to borders between labels.

    One specifies the max_offset (thickness) of the border. Then the offset is picked at random every time
    you call the transformer (offset is picked from the range 1:max_offset) for each axis and the boundary computed.
    One may use this scheme in order to make the network more robust against various thickness of borders in the
    ground truth (think of it as a boundary denoising scheme).

    Args:
        random_state: Random state for reproducibility.
        max_offset: Maximum offset for boundary thickness. Default: 10.
        ignore_index: Label to ignore in the output.
        append_label: If True, append original labels. Default: False.
        z_offset_scale: Scale factor for z-axis offsets. Default: 2.
    """

    def __init__(self, random_state, max_offset=10, ignore_index=None, append_label=False, z_offset_scale=2, **kwargs):
        super().__init__(ignore_index=ignore_index, append_label=append_label, aggregate_affinities=False)
        self.random_state = random_state
        self.offsets = tuple(range(1, max_offset + 1))
        self.z_offset_scale = z_offset_scale

    def get_kernels(self):
        rand_offset = self.random_state.choice(self.offsets)
        axis_ind = self.random_state.randint(3)
        # scale down z-affinities due to anisotropy
        if axis_ind == 2:
            rand_offset = max(1, rand_offset // self.z_offset_scale)

        rand_axis = self.AXES_TRANSPOSE[axis_ind]
        # return a single kernel
        return [self.create_kernel(rand_axis, rand_offset)]


class LabelToAffinities(AbstractLabelToBoundary):
    """Converts a given volumetric label array to binary mask corresponding to borders between labels.

    This can be seen as an affinity graph: https://arxiv.org/pdf/1706.00120.pdf
    One specifies the offsets (thickness) of the border. The boundary will be computed via the convolution operator.

    Args:
        offsets: List of offsets for boundary thickness.
        ignore_index: Label to ignore in the output.
        append_label: If True, append original labels. Default: False.
        aggregate_affinities: If True, aggregate affinities. Default: False.
        z_offsets: Offsets for z-axis, if different from xy offsets.
    """

    def __init__(
        self, offsets, ignore_index=None, append_label=False, aggregate_affinities=False, z_offsets=None, **kwargs
    ):
        super().__init__(
            ignore_index=ignore_index, append_label=append_label, aggregate_affinities=aggregate_affinities
        )

        assert isinstance(offsets, list) or isinstance(offsets, tuple), "offsets must be a list or a tuple"
        assert all(a > 0 for a in offsets), "'offsets' must be positive"
        assert len(set(offsets)) == len(offsets), "'offsets' must be unique"
        if z_offsets is not None:
            assert len(offsets) == len(z_offsets), "z_offsets length must be the same as the length of offsets"
        else:
            # if z_offsets is None just use the offsets for z-affinities
            z_offsets = list(offsets)
        self.z_offsets = z_offsets

        self.kernels = []
        # create kernel for every axis-offset pair
        for xy_offset, z_offset in zip(offsets, z_offsets, strict=True):
            for axis_ind, axis in enumerate(self.AXES_TRANSPOSE):
                final_offset = xy_offset
                if axis_ind == 2:
                    final_offset = z_offset
                # create kernels for a given offset in every direction
                self.kernels.append(self.create_kernel(axis, final_offset))

    def get_kernels(self):
        return self.kernels


class LabelToZAffinities(AbstractLabelToBoundary):
    """Converts a given volumetric label array to binary mask corresponding to borders between labels in Z-axis only.

    This can be seen as an affinity graph: https://arxiv.org/pdf/1706.00120.pdf
    One specifies the offsets (thickness) of the border. The boundary will be computed via the convolution operator.

    Args:
        offsets: List of offsets for boundary thickness.
        ignore_index: Label to ignore in the output.
        append_label: If True, append original labels. Default: False.
    """

    def __init__(self, offsets, ignore_index=None, append_label=False, **kwargs):
        super().__init__(ignore_index=ignore_index, append_label=append_label)

        assert isinstance(offsets, list) or isinstance(offsets, tuple), "offsets must be a list or a tuple"
        assert all(a > 0 for a in offsets), "'offsets must be positive"
        assert len(set(offsets)) == len(offsets), "'offsets' must be unique"

        self.kernels = []
        z_axis = self.AXES_TRANSPOSE[2]
        # create kernels
        for z_offset in offsets:
            self.kernels.append(self.create_kernel(z_axis, z_offset))

    def get_kernels(self):
        return self.kernels


class LabelToBoundaryAndAffinities:
    """Combines the StandardLabelToBoundary and LabelToAffinities.

    The hope is that training the network to predict both would improve the main task: boundary prediction.

    Args:
        xy_offsets: Offsets for XY axes.
        z_offsets: Offsets for Z axis.
        append_label: If True, append original labels. Default: False.
        ignore_index: Label to ignore in the output.
        mode: Boundary detection mode. Default: 'thick'.
        foreground: If True, include foreground mask. Default: False.
    """

    def __init__(
        self,
        xy_offsets: list,
        z_offsets: list,
        append_label: bool = False,
        ignore_index: int = None,
        mode: str = "thick",
        foreground: bool = False,
        **kwargs,
    ):
        # blur only StandardLabelToBoundary results; we don't want to blur the affinities
        self.l2b = StandardLabelToBoundary(ignore_index=ignore_index, mode=mode, foreground=foreground)
        self.l2a = LabelToAffinities(
            offsets=xy_offsets, z_offsets=z_offsets, append_label=append_label, ignore_index=ignore_index
        )

    def __call__(self, m: np.ndarray) -> np.ndarray:
        boundary = self.l2b(m)
        affinities = self.l2a(m)
        return np.concatenate((boundary, affinities), axis=0)


class LabelToMaskAndAffinities:
    """
    Similar to LabelToBoundaryAndAffinities but instead of computing the boundary we just compute the foreground
    in the first channel (everything greater than background is foreground).
    """

    def __init__(self, xy_offsets, z_offsets, append_label=False, background=0, ignore_index=None, **kwargs):
        self.background = background
        self.l2a = LabelToAffinities(
            offsets=xy_offsets, z_offsets=z_offsets, append_label=append_label, ignore_index=ignore_index
        )

    def __call__(self, m: np.ndarray) -> np.ndarray:
        mask = m > self.background
        mask = np.expand_dims(mask.astype(np.uint8), axis=0)
        affinities = self.l2a(m)
        return np.concatenate((mask, affinities), axis=0)


class Standardize:
    """Apply Z-score normalization to a given input tensor.

    Re-scales the values to be 0-mean and 1-std.

    Args:
        eps: Small value to prevent division by zero. Default: 1e-10.
        mean: Pre-computed mean value.
        std: Pre-computed standard deviation value.
        channelwise: If True, normalize per-channel. Default: False.
    """

    def __init__(self, eps: float = 1e-10, mean: float = None, std: float = None, channelwise: bool = False, **kwargs):
        if mean is not None or std is not None:
            assert mean is not None and std is not None
        self.mean = mean
        self.std = std
        self.eps = eps
        self.channelwise = channelwise

    def __call__(self, m: np.ndarray) -> np.ndarray:
        if self.mean is not None:
            mean, std = self.mean, self.std
        else:
            if self.channelwise:
                # normalize per-channel
                axes = list(range(m.ndim))
                # average across channels
                axes = tuple(axes[1:])
                mean = np.mean(m, axis=axes, keepdims=True)
                std = np.std(m, axis=axes, keepdims=True)
            else:
                mean = np.mean(m)
                std = np.std(m)

        return (m - mean) / np.clip(std, a_min=self.eps, a_max=None)


class PercentileNormalizer:
    """Apply percentile normalization to a given input tensor."""

    def __init__(self, pmin: float = 1.0, pmax: float = 99.6, channelwise: bool = False, eps: float = 1e-10, **kwargs):
        self.eps = eps
        self.pmin = pmin
        self.pmax = pmax
        self.channelwise = channelwise

    def __call__(self, m: np.ndarray) -> np.ndarray:
        if self.channelwise:
            axes = list(range(m.ndim))
            # average across channels
            axes = tuple(axes[1:])
            pmin = np.percentile(m, self.pmin, axis=axes, keepdims=True)
            pmax = np.percentile(m, self.pmax, axis=axes, keepdims=True)
        else:
            pmin = np.percentile(m, self.pmin)
            pmax = np.percentile(m, self.pmax)

        return (m - pmin) / (pmax - pmin + self.eps)


class Normalize:
    """Apply simple min-max scaling to a given input tensor.

    Shrinks the range of the data to a fixed range of [-1, 1] or in case of norm01==True to [0, 1].

    Args:
        min_value: Minimum value for clipping. Default: None (use min of the input array).
        max_value: Maximum value for clipping. Default: None (use max of the input array).
        norm01: If True, normalize to [0, 1] instead of [-1, 1]. Default: False.
        eps: Small value to prevent division by zero. Default: 1e-10.
    """

    def __init__(
        self, min_value: float = None, max_value: float = None, norm01: bool = False, eps: float = 1e-10, **kwargs
    ):
        if min_value is not None and max_value is not None:
            assert max_value > min_value
        self.min_value = min_value
        self.max_value = max_value
        self.norm01 = norm01
        self.eps = eps

    def __call__(self, m: np.ndarray) -> np.ndarray:
        if self.min_value is None:
            min_value = np.min(m)
        else:
            min_value = self.min_value

        if self.max_value is None:
            max_value = np.max(m)
        else:
            max_value = self.max_value

        # calculate norm_0_1 with min_value / max_value with the same dimension
        # in case of channelwise application
        norm_0_1 = (m - min_value) / (max_value - min_value + self.eps)

        if self.norm01:
            return np.clip(norm_0_1, 0, 1)
        else:
            return np.clip(2 * norm_0_1 - 1, -1, 1)


class AdditiveGaussianNoise:
    """Add Gaussian noise to a given input tensor."""

    def __init__(
        self,
        random_state: np.random.RandomState,
        scale: tuple[float, float] = (0.0, 1.0),
        execution_probability: float = 0.1,
        **kwargs,
    ):
        self.execution_probability = execution_probability
        self.random_state = random_state
        self.scale = scale

    def __call__(self, m: np.ndarray) -> np.ndarray:
        if self.random_state.uniform() < self.execution_probability:
            std = self.random_state.uniform(self.scale[0], self.scale[1])
            gaussian_noise = self.random_state.normal(0, std, size=m.shape)
            return m + gaussian_noise
        return m


class AdditivePoissonNoise:
    """Add Poisson noise to a given input tensor."""

    def __init__(
        self,
        random_state: np.random.RandomState,
        lam: tuple[float, float] = (0.0, 1.0),
        execution_probability: float = 0.1,
        **kwargs,
    ):
        self.execution_probability = execution_probability
        self.random_state = random_state
        self.lam = lam

    def __call__(self, m: np.ndarray) -> np.ndarray:
        if self.random_state.uniform() < self.execution_probability:
            lam = self.random_state.uniform(self.lam[0], self.lam[1])
            poisson_noise = self.random_state.poisson(lam, size=m.shape)
            return m + poisson_noise
        return m


class ToTensor:
    """
    Converts a given input numpy.ndarray into torch.Tensor.

    Args:
        expand_dims (bool): if True, adds a channel dimension to the input data
        dtype (np.dtype): the desired output data type
        normalize (bool): zero-one normalization of the input data
    """

    def __init__(self, expand_dims: bool, dtype: np.dtype = np.float32, normalize: bool = False, **kwargs):
        self.expand_dims = expand_dims
        self.dtype = dtype
        self.normalize = normalize

    def __call__(self, m: np.ndarray) -> torch.Tensor:
        assert m.ndim in [3, 4], "Supports only 3D (DxHxW) or 4D (CxDxHxW) images"
        # add channel dimension
        if self.expand_dims and m.ndim == 3:
            m = np.expand_dims(m, axis=0)

        if self.normalize:
            # avoid division by zero
            m = (m - np.min(m)) / (np.max(m) - np.min(m) + 1e-10)

        return torch.from_numpy(m.astype(dtype=self.dtype))


class Relabel:
    """Relabel a numpy array of labels into consecutive numbers.

    E.g. [10, 10, 0, 6, 6] -> [2, 2, 0, 1, 1]. Useful when one has an instance segmentation volume
    at hand and would like to create a one-hot-encoding for it. Without a consecutive labeling the task would be harder.

    Args:
        append_original: If True, append original labels. Default: False.
        run_cc: If True, run connected components. Default: True.
        ignore_label: Label to ignore.
    """

    def __init__(self, append_original: bool = False, run_cc: bool = True, ignore_label: int = None, **kwargs):
        self.append_original = append_original
        self.ignore_label = ignore_label
        self.run_cc = run_cc

        if ignore_label is not None:
            assert append_original, (
                "ignore_label present, so append_original must be true, so that one can localize the ignore region"
            )

    def __call__(self, m: np.ndarray) -> np.ndarray:
        orig = m
        if self.run_cc:
            # assign 0 to the ignore region
            m = measure.label(m, background=self.ignore_label)

        _, unique_labels = np.unique(m, return_inverse=True)
        result = unique_labels.reshape(m.shape)
        if self.append_original:
            result = np.stack([result, orig])
        return result


class Identity:
    def __init__(self, **kwargs):
        pass

    def __call__(self, m: np.ndarray) -> np.ndarray:
        return m


class RgbToLabel:
    """Convert a RGB image to a single channel label image."""

    def __call__(self, img: np.ndarray) -> np.ndarray:
        img = np.array(img)
        assert img.ndim == 3 and img.shape[2] == 3
        result = img[..., 0] * 65536 + img[..., 1] * 256 + img[..., 2]
        return result


class LabelToTensor:
    """Convert a given input numpy.ndarray label array into torch.Tensor of dtype int64."""

    def __call__(self, m: np.ndarray) -> torch.Tensor:
        m = np.array(m)
        return torch.from_numpy(m.astype(dtype="int64"))


class GaussianBlur3D:
    """Apply Gaussian blur to a given input tensor."""

    def __init__(self, sigma: tuple[float, float] = (0.1, 2.0), execution_probability: float = 0.5, **kwargs):
        self.sigma = sigma
        self.execution_probability = execution_probability

    def __call__(self, x: np.ndarray) -> np.ndarray:
        if random.random() < self.execution_probability:
            sigma = random.uniform(self.sigma[0], self.sigma[1])
            x = gaussian(x, sigma=sigma)
            return x
        return x


class Transformer:
    """Factory class for creating data augmentation pipelines."""

    def __init__(self, phase_config: dict, base_config: dict):
        self.phase_config = phase_config
        self.config_base = base_config
        self.seed = GLOBAL_RANDOM_STATE.randint(10000000)

    def raw_transform(self):
        return self._create_transform("raw")

    def label_transform(self):
        return self._create_transform("label")

    @staticmethod
    def _transformer_class(class_name):
        m = importlib.import_module("pytorch3dunet.augment.transforms")
        clazz = getattr(m, class_name)
        return clazz

    def _create_transform(self, name):
        assert name in self.phase_config, f"Could not find {name} transform"
        return Compose([self._create_augmentation(c) for c in self.phase_config[name]])

    def _create_augmentation(self, c):
        config = dict(self.config_base)
        config.update(c)
        config["random_state"] = np.random.RandomState(self.seed)
        aug_class = self._transformer_class(config["name"])
        return aug_class(**config)


def _recover_ignore_index(input, orig, ignore_index):
    if ignore_index is not None:
        mask = orig == ignore_index
        input[mask] = ignore_index

    return input



import numpy as np
import torch
import torch.nn as nn
from skimage import measure

from pytorch3dunet.augment.transforms import LabelToAffinities, StandardLabelToBoundary
from pytorch3dunet.unet3d.losses import (
    BCEDiceLoss,
    DiceLoss,
    GeneralizedDiceLoss,
    MaskingLossWrapper,
    SkipLastTargetChannelWrapper,
    WeightedSmoothL1Loss,
)
from pytorch3dunet.unet3d.metrics import (
    AdaptedRandError,
    BoundaryAdaptedRandError,
    BoundaryAveragePrecision,
    DiceCoefficient,
    MeanIoU,
)


def _compute_criterion(criterion, n_times=100):
    shape = [1, 0, 30, 30, 30]
    # channel size varies between 1 and 4
    results = []
    for C in range(1, 5):
        batch_shape = list(shape)
        batch_shape[1] = C
        batch_shape = tuple(batch_shape)
        results.append(_eval_criterion(criterion, batch_shape, n_times))

    return results


def _eval_criterion(criterion, batch_shape, n_times=100):
    with torch.no_grad():
        results = []
        # compute criterion n_times
        for _ in range(n_times):
            input = torch.rand(batch_shape)
            target = torch.zeros(batch_shape).random_(0, 2)
            output = criterion(input, target)
            results.append(output)

    return results


class TestCriterion:
    def test_dice_coefficient(self):
        results = _compute_criterion(DiceCoefficient())
        # check that all the coefficients belong to [0, 1]
        results = np.array(results)
        assert np.all(results > 0)
        assert np.all(results < 1)

    def test_mean_iou_simple(self):
        results = _compute_criterion(MeanIoU())
        # check that all of the coefficients belong to [0, 1]
        results = np.array(results)
        assert np.all(results > 0)
        assert np.all(results < 1)

    def test_mean_iou_multi_channel(self):
        criterion = MeanIoU()
        pred = torch.rand(10, 3, 10, 10, 10)
        target = pred > 0.5
        target = target.long()
        assert criterion(pred, target) == 1

    def test_mean_iou_multi_class(self):
        criterion = MeanIoU()
        n_classes = 5
        n_batch = 10
        pred = torch.rand(n_batch, n_classes, 10, 10, 10)
        target = torch.randint(0, n_classes, (n_batch, 10, 10, 10))
        mean_iou = criterion(pred, target)
        assert mean_iou >= 0

    def test_average_precision_synthethic_data(self):
        input = np.zeros((64, 200, 200), dtype=np.int32)
        for i in range(40, 200, 40):
            input[:, :, i : i + 2] = 1
        for i in range(40, 200, 40):
            input[:, i : i + 2, :] = 1
        for i in range(40, 64, 40):
            input[i : i + 2, :, :] = 1

        target = measure.label(np.logical_not(input).astype(np.int32), background=0)
        input = torch.tensor(input.reshape((1, 1) + input.shape))
        target = torch.tensor(target.reshape((1, 1) + target.shape))
        ap = BoundaryAveragePrecision()
        # expect perfect AP
        assert ap(input, target) == 1.0

    def test_average_precision_real_data(self, ovule_label):
        label = ovule_label[64:128, 64:128, 64:128]
        ltb = LabelToAffinities((1, 2, 4, 6), aggregate_affinities=True)
        pred = ltb(label)
        label = torch.tensor(label.reshape((1, 1) + label.shape).astype("int64"))
        pred = torch.tensor(np.expand_dims(pred, 0))
        ap = BoundaryAveragePrecision()
        assert ap(pred, label) > 0.5

    def test_adapted_rand_error(self, ovule_label):
        label = ovule_label[64:128, 64:128, 64:128].astype("int64")
        input = torch.tensor(label.reshape((1, 1) + label.shape))
        label = torch.tensor(label.reshape((1, 1) + label.shape))
        arand = AdaptedRandError()
        assert arand(input, label) == 0

    def test_adapted_rand_error_on_real_data(self, ovule_label):
        label = ovule_label[64:128, 64:128, 64:128].astype("int64")
        ltb = StandardLabelToBoundary()
        pred = ltb(label)
        label = torch.tensor(label.reshape((1, 1) + label.shape))
        pred = torch.tensor(np.expand_dims(pred, 0))
        arand = BoundaryAdaptedRandError(use_last_target=True)
        assert arand(pred, label) < 0.2

    def test_generalized_dice_loss(self):
        results = _compute_criterion(GeneralizedDiceLoss())
        # check that all of the coefficients belong to [0, 1]
        results = np.array(results)
        assert np.all(results > 0)
        assert np.all(results < 1)

    def test_dice_loss(self):
        results = _compute_criterion(DiceLoss())
        # check that all of the coefficients belong to [0, 1]
        results = np.array(results)
        assert np.all(results > 0)
        assert np.all(results < 1)

    def test_bce_dice_loss(self):
        results = _compute_criterion(BCEDiceLoss(1.0))
        results = np.array(results)
        assert np.all(results > 0)

    def test_ignore_index_loss(self):
        loss = MaskingLossWrapper(nn.BCEWithLogitsLoss(), ignore_index=-1)
        input = torch.rand((3, 3))
        input[1, 1] = 1.0
        input.requires_grad = True
        target = -1.0 * torch.ones((3, 3))
        target[1, 1] = 1.0
        output = loss(input, target)
        output.backward()

    def test_skip_last_target_channel(self):
        loss = SkipLastTargetChannelWrapper(nn.BCEWithLogitsLoss())
        input = torch.rand(1, 1, 3, 3, 3, requires_grad=True)
        target = torch.empty(1, 2, 3, 3, 3).random_(2)
        output = loss(input, target)
        output.backward()
        assert output.item() > 0

    def test_weighted_smooth_l1loss(self):
        loss_criterion = WeightedSmoothL1Loss(threshold=0.0, initial_weight=0.1)
        input = torch.randn(3, 16, 64, 64, 64, requires_grad=True)
        target = torch.randn(3, 16, 64, 64, 64)
        loss = loss_criterion(input, target)
        loss.backward()
        assert loss > 0



import importlib
import logging
import os
import shutil
import sys
from typing import Any

import numpy as np
import torch
from skimage.color import label2rgb
from torch import nn, optim
from torch.optim import Optimizer


def save_checkpoint(state: dict, is_best: bool, checkpoint_dir: str):
    """Saves model and training parameters at '{checkpoint_dir}/last_checkpoint.pytorch'.
    If is_best==True saves '{checkpoint_dir}/best_checkpoint.pytorch' as well.

    Args:
        state: contains model's state_dict, optimizer's state_dict, epoch
            and best evaluation metric value so far
        is_best: if True state contains the best model seen so far
        checkpoint_dir: directory where the checkpoint are to be saved
    """

    if not os.path.exists(checkpoint_dir):
        os.mkdir(checkpoint_dir)

    last_file_path = os.path.join(checkpoint_dir, "last_checkpoint.pytorch")
    torch.save(state, last_file_path)
    if is_best:
        best_file_path = os.path.join(checkpoint_dir, "best_checkpoint.pytorch")
        shutil.copyfile(last_file_path, best_file_path)


def load_checkpoint(
    checkpoint_path: str,
    model: nn.Module,
    optimizer: Optimizer = None,
    model_key: str = "model_state_dict",
    optimizer_key: str = "optimizer_state_dict",
) -> dict:
    """Loads model and training parameters from a given checkpoint_path
    If optimizer is provided, loads optimizer's state_dict of as well.

    Args:
        checkpoint_path: path to the checkpoint to be loaded
        model: model into which the parameters are to be copied
        optimizer: optimizer instance into which the parameters are to be copied
        model_key: key corresponding to the model state_dict in the checkpoint
        optimizer_key: key corresponding to the optimizer state_dict in the checkpoint

    Returns:
        state dict stored in the checkpoint
    """
    if not os.path.exists(checkpoint_path):
        raise OSError(f"Checkpoint '{checkpoint_path}' does not exist")

    state = torch.load(checkpoint_path, map_location="cpu")
    model.load_state_dict(state[model_key])

    if optimizer is not None:
        optimizer.load_state_dict(state[optimizer_key])

    return state


loggers = {}


def get_logger(name: str, level=logging.INFO) -> logging.Logger:
    """Initializes and returns a logger with the given name."""

    global loggers
    if loggers.get(name) is not None:
        return loggers[name]
    else:
        logger = logging.getLogger(name)
        logger.setLevel(level)
        # Logging to console
        stream_handler = logging.StreamHandler(sys.stdout)
        formatter = logging.Formatter("%(asctime)s [%(threadName)s] %(levelname)s %(name)s - %(message)s")
        stream_handler.setFormatter(formatter)
        logger.addHandler(stream_handler)

        loggers[name] = logger

        return logger


def get_number_of_learnable_parameters(model: nn.Module) -> int:
    """Returns the number of learnable parameters in the model."""
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


class RunningAverage:
    """Computes and stores the average"""

    def __init__(self):
        self.count = 0
        self.sum = 0
        self.avg = 0

    def update(self, value, n=1):
        self.count += n
        self.sum += value * n
        self.avg = self.sum / self.count


def number_of_features_per_level(init_channel_number: int, num_levels: int) -> list:
    """Computes the number of features at each level of the UNet"""
    return [init_channel_number * 2**k for k in range(num_levels)]


class TensorboardFormatter:
    """
    Tensorboard formatter converts a given batch of images (input/prediction/target)
    to a series of images that can be displayed in tensorboard.

    Args:
        skip_last_target (bool): if True, the last channel of the target image is skipped. This is useful for boundary
            based segmentation where the first channel is the boundary map and the second channel is the actual
            segmentation not used for training.
        log_channelwise (bool): if True, logs each channel of a multi-channel prediction, if False, takes the argmax
            over the channel dimension and logs a single label image.
    """

    def __init__(self, skip_last_target=False, log_channelwise=False):
        self.skip_last_target = skip_last_target
        self.log_channelwise = log_channelwise

    def __call__(self, name: str, batch: np.ndarray) -> list:
        """
        Transform a batch to a series of tuples of the form (tag, img), where `tag` corresponds to the image tag
        and `img` is the image itself.

        Args:
             name: one of 'inputs'/'targets'/'predictions'
             batch: 4D or 5D numpy array

        Returns:
            list[(str, np.ndarray)]: list of tuples of the form (tag, img)
        """

        def _check_img(tag_img):
            tag, img = tag_img

            assert img.ndim == 2 or img.ndim == 3, "Only 2D (HW) and 3D (CHW) images are accepted for display"

            if img.ndim == 2:
                img = np.expand_dims(img, axis=0)
            else:
                C = img.shape[0]
                assert C == 1 or C == 3, "Only (1, H, W) or (3, H, W) images are supported"

            return tag, img

        tagged_images = self._process_batch(name, batch)

        return list(map(_check_img, tagged_images))

    def _process_batch(self, name: str, batch: np.ndarray) -> list:
        if name == "targets" and self.skip_last_target:
            batch = batch[:, :-1, ...]

        tag_template = "{}/batch_{}/slice_{}"

        tagged_images = []

        if batch.ndim == 5:
            # NCDHW
            slice_idx = batch.shape[2] // 2  # get the middle slice
            for batch_idx in range(batch.shape[0]):
                if self.log_channelwise and name == "predictions":
                    tag_template = "{}/batch_{}/channel_{}/slice_{}"
                    for channel_idx in range(batch.shape[1]):
                        tag = tag_template.format(name, batch_idx, channel_idx, slice_idx)
                        img = batch[batch_idx, channel_idx, slice_idx, ...]
                        tagged_images.append((tag, self._normalize_img(img)))
                else:
                    tag = tag_template.format(name, batch_idx, slice_idx)
                    if name in ["predictions", "targets"]:
                        # for single channel predictions, just log the image
                        if batch.shape[1] == 1:
                            img = batch[batch_idx, :, slice_idx, ...]
                            tagged_images.append((tag, self._normalize_img(img)))
                        else:
                            # predictions are probabilities so convert to label image
                            img = batch[batch_idx].argmax(axis=0)
                            # take the middle slice
                            img = img[slice_idx, ...]
                            # convert to label image
                            img = label2rgb(img)
                            img = img.transpose(2, 0, 1)
                            tagged_images.append((tag, img))
                    else:
                        # handle input images
                        if batch.shape[1] in [1, 3]:
                            # if single channel or RGB image, log directly
                            img = batch[batch_idx, :, slice_idx, ...]
                            tagged_images.append((tag, self._normalize_img(img)))
                        else:
                            # log channelwise
                            tag_template = "{}/batch_{}/channel_{}/slice_{}"
                            for channel_idx in range(batch.shape[1]):
                                tag = tag_template.format(name, batch_idx, channel_idx, slice_idx)
                                img = batch[batch_idx, channel_idx, slice_idx, ...]
                                tagged_images.append((tag, self._normalize_img(img)))

        else:
            # batch has no channel dim: NDHW
            slice_idx = batch.shape[1] // 2  # get the middle slice
            for batch_idx in range(batch.shape[0]):
                tag = tag_template.format(name, batch_idx, slice_idx)
                img = batch[batch_idx, slice_idx, ...]
                # this is target segmentation so convert to label image
                lbl = label2rgb(img)
                lbl = lbl.transpose(2, 0, 1)
                tagged_images.append((tag, lbl))

        return tagged_images

    @staticmethod
    def _normalize_img(img: np.ndarray) -> np.ndarray:
        return np.nan_to_num((img - np.min(img)) / np.ptp(img))


def convert_to_numpy(*inputs):
    """
    Coverts input tensors to numpy ndarrays

    Args:
        inputs (iteable of torch.Tensor): torch tensor

    Returns:
        tuple of ndarrays
    """

    def _to_numpy(i):
        assert isinstance(i, torch.Tensor), "Expected input to be torch.Tensor"
        return i.detach().cpu().numpy()

    return (_to_numpy(i) for i in inputs)


def create_optimizer(optimizer_config: dict, model: nn.Module) -> Optimizer:
    optim_name = optimizer_config.get("name", "Adam")
    # common optimizer settings
    learning_rate = optimizer_config.get("learning_rate", 1e-3)
    weight_decay = optimizer_config.get("weight_decay", 0)

    # grab optimizer specific settings and init
    # optimizer
    if optim_name == "Adadelta":
        rho = optimizer_config.get("rho", 0.9)
        optimizer = optim.Adadelta(model.parameters(), lr=learning_rate, rho=rho, weight_decay=weight_decay)
    elif optim_name == "Adagrad":
        lr_decay = optimizer_config.get("lr_decay", 0)
        optimizer = optim.Adagrad(model.parameters(), lr=learning_rate, lr_decay=lr_decay, weight_decay=weight_decay)
    elif optim_name == "AdamW":
        betas = tuple(optimizer_config.get("betas", (0.9, 0.999)))
        optimizer = optim.AdamW(model.parameters(), lr=learning_rate, betas=betas, weight_decay=weight_decay)
    elif optim_name == "SparseAdam":
        betas = tuple(optimizer_config.get("betas", (0.9, 0.999)))
        optimizer = optim.SparseAdam(model.parameters(), lr=learning_rate, betas=betas)
    elif optim_name == "Adamax":
        betas = tuple(optimizer_config.get("betas", (0.9, 0.999)))
        optimizer = optim.Adamax(model.parameters(), lr=learning_rate, betas=betas, weight_decay=weight_decay)
    elif optim_name == "LBFGS":
        max_iter = optimizer_config.get("max_iter", 20)
        max_eval = optimizer_config.get("max_eval", None)
        tolerance_grad = optimizer_config.get("tolerance_grad", 1e-7)
        tolerance_change = optimizer_config.get("tolerance_change", 1e-9)
        history_size = optimizer_config.get("history_size", 100)
        optimizer = optim.LBFGS(
            model.parameters(),
            lr=learning_rate,
            max_iter=max_iter,
            max_eval=max_eval,
            tolerance_grad=tolerance_grad,
            tolerance_change=tolerance_change,
            history_size=history_size,
        )
    elif optim_name == "NAdam":
        betas = tuple(optimizer_config.get("betas", (0.9, 0.999)))
        momentum_decay = optimizer_config.get("momentum_decay", 4e-3)
        optimizer = optim.NAdam(
            model.parameters(), lr=learning_rate, betas=betas, momentum_decay=momentum_decay, weight_decay=weight_decay
        )
    elif optim_name == "RAdam":
        betas = tuple(optimizer_config.get("betas", (0.9, 0.999)))
        optimizer = optim.RAdam(model.parameters(), lr=learning_rate, betas=betas, weight_decay=weight_decay)
    elif optim_name == "RMSprop":
        alpha = optimizer_config.get("alpha", 0.99)
        optimizer = optim.RMSprop(model.parameters(), lr=learning_rate, alpha=alpha, weight_decay=weight_decay)
    elif optim_name == "Rprop":
        etas = optimizer_config.get("etas", (0.5, 1.2))
        step_sizes = optimizer_config.get("step_sizes", (1e-6, 50))
        optimizer = optim.Rprop(model.parameters(), lr=learning_rate, etas=etas, step_sizes=step_sizes)
    elif optim_name == "SGD":
        momentum = optimizer_config.get("momentum", 0)
        dampening = optimizer_config.get("dampening", 0)
        nesterov = optimizer_config.get("nesterov", False)
        optimizer = optim.SGD(
            model.parameters(),
            lr=learning_rate,
            momentum=momentum,
            dampening=dampening,
            nesterov=nesterov,
            weight_decay=weight_decay,
        )
    else:  # Adam is default
        betas = tuple(optimizer_config.get("betas", (0.9, 0.999)))
        optimizer = optim.Adam(model.parameters(), lr=learning_rate, betas=betas, weight_decay=weight_decay)

    return optimizer


def create_lr_scheduler(lr_config: dict, optimizer: Optimizer) -> Any | None:
    """Creates a learning rate scheduler"""
    if lr_config is None:
        return None
    class_name = lr_config.pop("name")
    m = importlib.import_module("torch.optim.lr_scheduler")
    clazz = getattr(m, class_name)
    # add optimizer to the config
    lr_config["optimizer"] = optimizer
    return clazz(**lr_config)


def get_class(class_name: str, modules: list[str]) -> type:
    """Helper function which searches for a class in the given list of modules and returns it."""
    for module in modules:
        m = importlib.import_module(module)
        clazz = getattr(m, class_name, None)
        if clazz is not None:
            return clazz
    raise RuntimeError(f"Unsupported dataset class: {class_name}")



import os
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from typing import Any

import numpy as np
import torch
import torch.nn as nn
from torch.optim.lr_scheduler import ReduceLROnPlateau
from torch.utils.tensorboard import SummaryWriter
from tqdm import tqdm

from pytorch3dunet.datasets.utils import get_train_loaders
from pytorch3dunet.unet3d.config import TorchDevice
from pytorch3dunet.unet3d.losses import get_loss_criterion
from pytorch3dunet.unet3d.metrics import get_evaluation_metric
from pytorch3dunet.unet3d.model import get_model, is_model_2d
from pytorch3dunet.unet3d.utils import (
    RunningAverage,
    TensorboardFormatter,
    create_lr_scheduler,
    create_optimizer,
    get_logger,
    get_number_of_learnable_parameters,
    load_checkpoint,
    save_checkpoint,
)

logger = get_logger("UNetTrainer")


def create_trainer(config: dict) -> "UNetTrainer":
    # Create the model
    model = get_model(config["model"])

    device = config.get("device", None)
    assert device, "Device not specified in the config file and could not be inferred automatically"
    logger.info(f"Using device: {device}")
    model.to(device)

    # Log the number of learnable parameters
    logger.info(f"Number of learnable params {get_number_of_learnable_parameters(model)}")

    # Create loss criterion
    loss_criterion = get_loss_criterion(config)
    # Create evaluation metric
    eval_criterion = get_evaluation_metric(config)

    # Create data loaders
    loaders = get_train_loaders(config)

    # Create the optimizer
    optimizer = create_optimizer(config["optimizer"], model)

    # Create learning rate adjustment strategy
    lr_scheduler = create_lr_scheduler(config.get("lr_scheduler", None), optimizer)

    trainer_config = config["trainer"]
    # Create tensorboard formatter
    tensorboard_formatter_config = trainer_config.pop("tensorboard_formatter", {})
    tensorboard_formatter = TensorboardFormatter(**tensorboard_formatter_config)
    # Create trainer
    resume = trainer_config.pop("resume", None)
    pre_trained = trainer_config.pop("pre_trained", None)

    return UNetTrainer(
        model=model,
        optimizer=optimizer,
        lr_scheduler=lr_scheduler,
        loss_criterion=loss_criterion,
        eval_criterion=eval_criterion,
        loaders=loaders,
        tensorboard_formatter=tensorboard_formatter,
        resume=resume,
        pre_trained=pre_trained,
        device=device,
        **trainer_config,
    )


def _split_and_move_to_device(t: Any, device: TorchDevice) -> tuple[torch.Tensor, torch.Tensor]:
    def _move_to_device(input, device):
        if isinstance(input, tuple | list):
            return tuple([_move_to_device(x, device) for x in input])
        else:
            # send batch to device, see: https://docs.pytorch.org/tutorials/intermediate/pinmem_nonblock.html
            return input.to(device, non_blocking=True)

    input, target = _move_to_device(t, device)
    return input, target


class UNetTrainer:
    """UNet trainer.

    Args:
        model: UNet 3D model to be trained.
        optimizer: Optimizer used for training.
        lr_scheduler: Learning rate scheduler. Note that lr_scheduler.step() is invoked after every validation
            step (i.e. validate_after_iters) not after every epoch. So e.g. if one uses StepLR with step_size=30
            the learning rate will be adjusted after every 30 * validate_after_iters iterations.
        loss_criterion: Loss function.
        eval_criterion: Used to compute training/validation metric (such as Dice, IoU, AP or Rand score).
            Saving the best checkpoint is based on the result of this function on the validation set.
        loaders: Dictionary with 'train' and 'val' data loaders.
        checkpoint_dir: Directory for saving checkpoints and tensorboard logs.
        max_num_epochs: Maximum number of epochs.
        max_num_iterations: Maximum number of iterations.
        validate_after_iters: Validate after that many iterations. Default: 200.
        log_after_iters: Number of iterations before logging to tensorboard. Default: 100.
        validate_iters: Number of validation iterations. If None validate on the whole validation set. Default: None.
        num_iterations: Useful when loading the model from the checkpoint. Default: 1.
        num_epoch: Useful when loading the model from the checkpoint. Default: 0.
        eval_score_higher_is_better: If True higher eval scores are considered better. Default: True.
        tensorboard_formatter: Converts a given batch of input/output/target image to a series of images
            that can be displayed in tensorboard. Default: None.
        skip_train_validation: If True eval_criterion is not evaluated on the training set (used when
            evaluation is expensive). Default: False.
        resume: Path to the checkpoint to be resumed. Default: None.
        pre_trained: Path to the pre-trained model. Default: None.
        max_val_images: Maximum number of images to log during validation. Default: 100.
        device: Device to use for training (CPU, CUDA, MPS). Default: None.
    """

    def __init__(
        self,
        model,
        optimizer,
        lr_scheduler,
        loss_criterion,
        eval_criterion,
        loaders,
        checkpoint_dir,
        max_num_epochs,
        max_num_iterations,
        validate_after_iters=200,
        log_after_iters=100,
        validate_iters=None,
        num_iterations=1,
        num_epoch=0,
        eval_score_higher_is_better=True,
        tensorboard_formatter=None,
        skip_train_validation=False,
        resume=None,
        pre_trained=None,
        max_val_images=100,
        device: TorchDevice | None = None,
    ):
        self.max_val_images = max_val_images
        self.model = model
        self.optimizer = optimizer
        self.scheduler = lr_scheduler
        self.loss_criterion = loss_criterion
        self.eval_criterion = eval_criterion
        self.loaders = loaders
        self.checkpoint_dir = checkpoint_dir
        self.max_num_epochs = max_num_epochs
        self.max_num_iterations = max_num_iterations
        self.validate_after_iters = validate_after_iters
        self.log_after_iters = log_after_iters
        self.validate_iters = validate_iters
        self.eval_score_higher_is_better = eval_score_higher_is_better
        assert device, "Device must be specified"
        self.device = device

        logger.info(model)
        logger.info(f"eval_score_higher_is_better: {eval_score_higher_is_better}")
        # initialize the best_eval_score
        if eval_score_higher_is_better:
            self.best_eval_score = float("-inf")
        else:
            self.best_eval_score = float("+inf")

        self.writer = SummaryWriter(
            log_dir=os.path.join(checkpoint_dir, "logs", datetime.now().strftime("%Y-%m-%d_%H-%M-%S"))
        )

        assert tensorboard_formatter is not None, "TensorboardFormatter must be provided"
        self.tensorboard_formatter = tensorboard_formatter

        self.num_iterations = num_iterations
        self.num_epochs = num_epoch
        self.skip_train_validation = skip_train_validation

        if resume is not None:
            logger.info(f"Loading checkpoint '{resume}'...")
            state = load_checkpoint(resume, self.model, self.optimizer)
            logger.info(
                f"Checkpoint loaded from '{resume}'. Epoch: {state['num_epochs']}.  Iteration: {state['num_iterations']}. "
                f"Best val score: {state['best_eval_score']}."
            )
            self.best_eval_score = state["best_eval_score"]
            self.num_iterations = state["num_iterations"]
            self.num_epochs = state["num_epochs"]
            self.checkpoint_dir = os.path.split(resume)[0]
        elif pre_trained is not None:
            logger.info(f"Logging pre-trained model from '{pre_trained}'...")
            load_checkpoint(pre_trained, self.model, None)
            if not self.checkpoint_dir:
                self.checkpoint_dir = os.path.split(pre_trained)[0]

        # use DataParallel if more than 1 GPU available
        if device == TorchDevice.CUDA and torch.cuda.device_count() > 1:
            self.model = nn.DataParallel(self.model)
            logger.info(f"Using {torch.cuda.device_count()} GPUs for training")

    def fit(self):
        for _ in range(self.num_epochs, self.max_num_epochs):
            # train for one epoch
            should_terminate = self.train()

            if should_terminate:
                logger.info("Stopping criterion is satisfied. Finishing training")
                return

            self.num_epochs += 1
        logger.info(f"Reached maximum number of epochs: {self.max_num_epochs}. Finishing training...")

    def train(self):
        """Trains the model for 1 epoch.

        Returns:
            True if the training should be terminated immediately, False otherwise
        """
        train_losses = RunningAverage()
        train_eval_scores = RunningAverage()

        # sets the model in training mode
        self.model.train()

        for t in self.loaders["train"]:
            logger.info(
                f"Training iteration [{self.num_iterations}/{self.max_num_iterations}]. "
                f"Epoch [{self.num_epochs}/{self.max_num_epochs - 1}]"
            )

            input, target = _split_and_move_to_device(t, self.device)

            output, loss = self._forward_pass(input, target)

            train_losses.update(loss.item(), self._batch_size(input))

            # compute gradients and update parameters
            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()

            if self.num_iterations % self.validate_after_iters == 0:
                # set the model in eval mode
                self.model.eval()
                # evaluate on validation set
                eval_score = self.validate()
                # set the model back to training mode
                self.model.train()

                # adjust learning rate if necessary
                if isinstance(self.scheduler, ReduceLROnPlateau):
                    self.scheduler.step(eval_score)
                elif self.scheduler is not None:
                    self.scheduler.step()

                # log current learning rate in tensorboard
                self._log_lr()
                # remember the best validation metric
                is_best = self._is_best_eval_score(eval_score)

                # save checkpoint
                self._save_checkpoint(is_best)

            if self.num_iterations % self.log_after_iters == 0:
                # compute eval criterion
                if not self.skip_train_validation:
                    eval_score = self.eval_criterion(output, target)
                    train_eval_scores.update(eval_score.item(), self._batch_size(input))

                logger.info(f"Training stats. Loss: {train_losses.avg}. Evaluation score: {train_eval_scores.avg}")
                self._log_stats("train", train_losses.avg, train_eval_scores.avg)
                self._log_images(
                    input.detach().cpu().numpy(), target.detach().cpu().numpy(), output.detach().cpu().numpy(), "train_"
                )

            if self.should_stop():
                return True

            self.num_iterations += 1

        return False

    def should_stop(self):
        """Check if training should be terminated.

        Training will terminate if maximum number of iterations is exceeded or the learning rate drops below
        some predefined threshold (1e-6 in our case).

        Returns:
            True if training should stop, False otherwise.
        """
        if self.max_num_iterations < self.num_iterations:
            logger.info(f"Maximum number of iterations {self.max_num_iterations} exceeded.")
            return True

        min_lr = 1e-6
        lr = self.optimizer.param_groups[0]["lr"]
        if lr < min_lr:
            logger.info(f"Learning rate below the minimum {min_lr}.")
            return True

        return False

    def validate(self):
        logger.info("Validating...")

        val_losses = RunningAverage()
        val_scores = RunningAverage()

        with torch.no_grad():
            # select indices of validation samples to log
            rs = np.random.RandomState(42)
            if len(self.loaders["val"]) <= self.max_val_images:
                indices = list(range(len(self.loaders["val"])))
            else:
                indices = rs.choice(len(self.loaders["val"]), size=self.max_val_images, replace=False)

            images_for_logging = []
            for i, t in enumerate(tqdm(self.loaders["val"])):
                input, target = _split_and_move_to_device(t, self.device)

                output, loss = self._forward_pass(input, target)
                val_losses.update(loss.item(), self._batch_size(input))
                eval_score = self.eval_criterion(output, target)
                val_scores.update(eval_score.item(), self._batch_size(input))

                # save val images for logging
                if i in indices:
                    imgs = (input.cpu().numpy(), target.cpu().numpy(), output.cpu().numpy())
                    images_for_logging.append(imgs + (i,))

                if self.validate_iters is not None and self.validate_iters <= i:
                    # stop validation
                    break

            # log images in a separate thread
            with ThreadPoolExecutor() as executor:
                for input, target, output, i in images_for_logging:
                    executor.submit(self._log_images, input, target, output, f"val_{i}_")

            logger.info(f"Validation finished. Loss: {val_losses.avg}. Evaluation score: {val_scores.avg}")
            self._log_stats("val", val_losses.avg, val_scores.avg)
            return val_scores.avg

    def _forward_pass(self, inp: torch.Tensor, target: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        if is_model_2d(self.model):
            # remove the singleton z-dimension from the input
            inp = torch.squeeze(inp, dim=-3)
            # forward pass
            output, logits = self.model(inp, return_logits=True)
            # add the singleton z-dimension to the output
            output = torch.unsqueeze(output, dim=-3)
            logits = torch.unsqueeze(logits, dim=-3)
        else:
            # forward pass
            output, logits = self.model(inp, return_logits=True)

        # always compute the loss using logits
        loss = self.loss_criterion(logits, target)

        # return probabilities and loss
        return output, loss

    def _is_best_eval_score(self, eval_score: float) -> bool:
        if self.eval_score_higher_is_better:
            is_best = eval_score > self.best_eval_score
        else:
            is_best = eval_score < self.best_eval_score

        if is_best:
            logger.info(f"Saving new best evaluation metric: {eval_score}")
            self.best_eval_score = eval_score

        return is_best

    def _save_checkpoint(self, is_best: bool):
        # remove `module` prefix from layer names when using `nn.DataParallel`
        # see: https://discuss.pytorch.org/t/solved-keyerror-unexpected-key-module-encoder-embedding-weight-in-state-dict/1686/20
        if isinstance(self.model, nn.DataParallel):
            state_dict = self.model.module.state_dict()
        else:
            state_dict = self.model.state_dict()

        last_file_path = os.path.join(self.checkpoint_dir, "last_checkpoint.pytorch")
        logger.info(f"Saving checkpoint to '{last_file_path}'")

        save_checkpoint(
            {
                "num_epochs": self.num_epochs + 1,
                "num_iterations": self.num_iterations,
                "model_state_dict": state_dict,
                "best_eval_score": self.best_eval_score,
                "optimizer_state_dict": self.optimizer.state_dict(),
            },
            is_best,
            checkpoint_dir=self.checkpoint_dir,
        )

    def _log_lr(self):
        lr = self.optimizer.param_groups[0]["lr"]
        self.writer.add_scalar("learning_rate", lr, self.num_iterations)

    def _log_stats(self, phase: str, loss_avg: float, eval_score_avg: float):
        tag_value = {f"{phase}_loss_avg": loss_avg, f"{phase}_eval_score_avg": eval_score_avg}

        for tag, value in tag_value.items():
            self.writer.add_scalar(tag, value, self.num_iterations)

    def _log_params(self):
        logger.info("Logging model parameters and gradients")
        for name, value in self.model.named_parameters():
            self.writer.add_histogram(name, value.data.cpu().numpy(), self.num_iterations)
            self.writer.add_histogram(name + "/grad", value.grad.data.cpu().numpy(), self.num_iterations)

    def _log_images(self, input: np.ndarray, target: np.ndarray, prediction: np.ndarray, prefix: str):
        inputs_map = {"inputs": input, "targets": target, "predictions": prediction}
        img_sources = {}
        for name, batch in inputs_map.items():
            if isinstance(batch, list | tuple):
                for i, b in enumerate(batch):
                    img_sources[f"{name}{i}"] = b
            else:
                img_sources[name] = batch

        for name, batch in img_sources.items():
            for tag, image in self.tensorboard_formatter(name, batch):
                self.writer.add_image(prefix + tag, image, self.num_iterations)

    @staticmethod
    def _batch_size(input: torch.Tensor) -> int:
        if isinstance(input, list) or isinstance(input, tuple):
            return input[0].size(0)
        else:
            return input.size(0)



import os
from tempfile import NamedTemporaryFile

import h5py
import numpy as np

from pytorch3dunet.datasets.utils import get_train_loaders
from pytorch3dunet.unet3d.losses import get_loss_criterion
from pytorch3dunet.unet3d.metrics import get_evaluation_metric
from pytorch3dunet.unet3d.model import get_model
from pytorch3dunet.unet3d.trainer import UNetTrainer
from pytorch3dunet.unet3d.utils import TensorboardFormatter, create_lr_scheduler, create_optimizer


class TestUNet3DTrainer:
    def test_ce_loss(self, tmpdir, capsys, train_config):
        with capsys.disabled():
            assert_train_save_load(tmpdir, train_config, "CrossEntropyLoss", "MeanIoU", "UNet3D")

    def test_wce_loss(self, tmpdir, capsys, train_config):
        with capsys.disabled():
            assert_train_save_load(tmpdir, train_config, "WeightedCrossEntropyLoss", "MeanIoU", "UNet3D")

    def test_bce_loss(self, tmpdir, capsys, train_config):
        with capsys.disabled():
            assert_train_save_load(tmpdir, train_config, "BCEWithLogitsLoss", "DiceCoefficient", "UNet3D")

    def test_dice_loss(self, tmpdir, capsys, train_config):
        with capsys.disabled():
            assert_train_save_load(tmpdir, train_config, "DiceLoss", "MeanIoU", "UNet3D")

    def test_residual_unet(self, tmpdir, capsys, train_config):
        with capsys.disabled():
            assert_train_save_load(tmpdir, train_config, "CrossEntropyLoss", "MeanIoU", "ResidualUNet3D")

    def test_2d_unet(self, tmpdir, capsys, train_config_2d):
        with capsys.disabled():
            assert_train_save_load(
                tmpdir, train_config_2d, "CrossEntropyLoss", "MeanIoU", "UNet2D", shape=(3, 1, 128, 128)
            )


def assert_train_save_load(tmpdir, train_config, loss, val_metric, model, weight_map=False, shape=(3, 64, 64, 64)):
    max_num_epochs = train_config["trainer"]["max_num_epochs"]
    log_after_iters = train_config["trainer"]["log_after_iters"]
    validate_after_iters = train_config["trainer"]["validate_after_iters"]
    max_num_iterations = train_config["trainer"]["max_num_iterations"]

    trainer = _train_save_load(tmpdir, train_config, loss, val_metric, model, weight_map, shape)

    assert trainer.num_iterations == max_num_iterations
    assert trainer.max_num_epochs == max_num_epochs
    assert trainer.log_after_iters == log_after_iters
    assert trainer.validate_after_iters == validate_after_iters
    assert trainer.max_num_iterations == max_num_iterations


def _train_save_load(tmpdir, train_config, loss, val_metric, model, weight_map, shape):
    binary_loss = loss in ["BCEWithLogitsLoss", "DiceLoss", "BCEDiceLoss", "GeneralizedDiceLoss"]

    train_config["model"]["name"] = model
    train_config.update(
        {
            # get device to train on
            "loss": {"name": loss, "weight": np.random.rand(2).astype(np.float32), "pos_weight": 3.0},
            "eval_metric": {"name": val_metric},
        }
    )
    train_config["model"]["final_sigmoid"] = binary_loss

    if weight_map:
        train_config["loaders"]["weight_internal_path"] = "weight_map"

    loss_criterion = get_loss_criterion(train_config)
    eval_criterion = get_evaluation_metric(train_config)
    model = get_model(train_config["model"])
    model = model.to(train_config["device"])

    if loss in ["BCEWithLogitsLoss"]:
        label_dtype = "float32"
        train_config["loaders"]["train"]["transformer"]["label"][0]["dtype"] = label_dtype
        train_config["loaders"]["val"]["transformer"]["label"][0]["dtype"] = label_dtype

    train = _create_random_dataset(shape, binary_loss)
    val = _create_random_dataset(shape, binary_loss)
    train_config["loaders"]["train"]["file_paths"] = [train]
    train_config["loaders"]["val"]["file_paths"] = [val]

    loaders = get_train_loaders(train_config)

    optimizer = create_optimizer(train_config["optimizer"], model)
    lr_scheduler = create_lr_scheduler(train_config.get("lr_scheduler", None), optimizer)

    formatter = TensorboardFormatter()
    trainer = UNetTrainer(
        model,
        optimizer,
        lr_scheduler,
        loss_criterion,
        eval_criterion,
        loaders,
        tmpdir,
        max_num_epochs=train_config["trainer"]["max_num_epochs"],
        max_num_iterations=train_config["trainer"]["max_num_iterations"],
        validate_after_iters=train_config["trainer"]["log_after_iters"],
        log_after_iters=train_config["trainer"]["log_after_iters"],
        tensorboard_formatter=formatter,
        device=train_config["device"],
    )
    trainer.fit()
    # test loading the trainer from the checkpoint
    trainer = UNetTrainer(
        model,
        optimizer,
        lr_scheduler,
        loss_criterion,
        eval_criterion,
        loaders,
        tmpdir,
        max_num_epochs=train_config["trainer"]["max_num_epochs"],
        max_num_iterations=train_config["trainer"]["max_num_iterations"],
        validate_after_iters=train_config["trainer"]["log_after_iters"],
        log_after_iters=train_config["trainer"]["log_after_iters"],
        tensorboard_formatter=formatter,
        resume=os.path.join(tmpdir, "last_checkpoint.pytorch"),
        device=train_config["device"],
    )
    return trainer


def _create_random_dataset(shape, channel_per_class):
    tmp = NamedTemporaryFile(delete=False)

    with h5py.File(tmp.name, "w") as f:
        l_shape = w_shape = shape
        # make sure that label and weight tensors are 3D
        if len(shape) == 4:
            l_shape = shape[1:]
            w_shape = shape[1:]

        if channel_per_class:
            l_shape = (2,) + l_shape

        f.create_dataset("raw", data=np.random.rand(*shape))
        f.create_dataset("label", data=np.random.randint(0, 2, l_shape))
        f.create_dataset("weight_map", data=np.random.rand(*w_shape))

    return tmp.name



import importlib

import numpy as np
import torch
from skimage import measure
from skimage.metrics import adapted_rand_error, mean_squared_error, peak_signal_noise_ratio

from pytorch3dunet.unet3d.losses import compute_per_channel_dice
from pytorch3dunet.unet3d.seg_metrics import Accuracy, AveragePrecision
from pytorch3dunet.unet3d.utils import convert_to_numpy, get_logger

logger = get_logger("EvalMetric")


class DiceCoefficient:
    """Computes Dice Coefficient.
    Generalized to multiple channels by computing per-channel Dice Score
    (as described in https://arxiv.org/pdf/1707.03237.pdf) and then simply taking the average.
    Input is expected to be probabilities instead of logits.
    """

    def __init__(self, epsilon=1e-6, **kwargs):
        self.epsilon = epsilon

    def __call__(self, input, target):
        # Average across channels in order to get the final score
        return torch.mean(compute_per_channel_dice(input, target, epsilon=self.epsilon))


class MeanIoU:
    """
    Computes IoU for each class separately and then averages over all classes.

    Args:
        skip_background (bool): if True, background class (i.e. 0-label) will be skipped when computing IoU
    """

    def __init__(self, skip_background=True, **kwargs):
        self.skip_background = skip_background

    def __call__(self, input, target):
        """Compute mean IoU.

        Args:
            input: 5D probability maps torch float tensor (NxCxDxHxW).
            target: 4D or 5D ground truth torch tensor. 4D (NxDxHxW) tensor will be expanded to 5D as one-hot.

        Returns:
            Intersection over union averaged over all channels.
        """
        assert input.dim() == 5

        n_classes = input.size(1)

        if target.dim() == 4:
            # convert input to segmentation
            input = input.argmax(dim=1)

        assert input.size() == target.size()

        per_batch_iou = []
        for _input, _target in zip(input, target, strict=True):
            # convert target to byte
            _target = _target.byte()
            per_channel_iou = []
            start_idx = 0
            # skip background only if target is 4D; for channel-wise computation (i.e. if target is 5D) we need to include it
            if self.skip_background and target.dim() == 4:
                start_idx = 1

            for c in range(start_idx, n_classes):
                if target.dim() == 5:
                    iou = self._jaccard_index(_input[c] > 0.5, _target[c])
                    per_channel_iou.append(iou)
                else:
                    iou = self._jaccard_index(_input == c, _target == c)
                    per_channel_iou.append(iou)

            assert per_channel_iou, "All channels were ignored from the computation"
            mean_iou = torch.tensor(per_channel_iou).mean()
            per_batch_iou.append(mean_iou)

        return torch.tensor(per_batch_iou).mean()

    def _jaccard_index(self, prediction, target):
        """Computes IoU for a given target and prediction tensors.

        Args:
            prediction: Predicted binary mask.
            target: Ground truth binary mask.

        Returns:
            Intersection over union score.
        """
        epsilon = 1e-8
        intersection = torch.logical_and(target, prediction).sum()
        union = torch.logical_or(target, prediction).sum()
        return (intersection + epsilon) / (union + epsilon)


class AdaptedRandError:
    """
    A functor which computes an Adapted Rand error as defined by the SNEMI3D contest
    (http://brainiac2.mit.edu/SNEMI3D/evaluation).

    This is a generic implementation which takes the input, converts it to the segmentation image (see `input_to_segm()`)
    and then computes the ARand between the segmentation and the ground truth target. Depending on one's use case
    it's enough to extend this class and implement the `input_to_segm` method.

    Args:
        use_last_target (bool): if true, use the last channel from the target to compute the ARand, otherwise the first.
    """

    def __init__(self, use_last_target=False, ignore_index=None, **kwargs):
        self.use_last_target = use_last_target
        self.ignore_index = ignore_index

    def __call__(self, input, target):
        """
        Compute ARand Error for each input, target pair in the batch and return the mean value.

        Args:
            input (torch.tensor):  5D (NCDHW) output from the network
            target (torch.tensor): 5D (NCDHW) ground truth segmentation

        Returns:
            average ARand Error across the batch
        """

        # converts input and target to numpy arrays
        input, target = convert_to_numpy(input, target)
        if self.use_last_target:
            target = target[:, -1, ...]  # 4D
        else:
            # use 1st target channel
            target = target[:, 0, ...]  # 4D

        # ensure target is of integer type
        target = target.astype(np.int32)

        if self.ignore_index is not None:
            target[target == self.ignore_index] = 0

        per_batch_arand = []
        for _input, _target in zip(input, target, strict=True):
            if np.all(
                _target == _target.flat[0]
            ):  # skip ARand eval if there is only one label in the patch due to zero-division
                logger.info("Skipping ARandError computation: only 1 label present in the ground truth")
                per_batch_arand.append(0.0)
                continue

            # convert _input to segmentation CDHW
            segm = self.input_to_segm(_input)
            assert segm.ndim == 4

            # compute per channel arand and return the minimum value
            per_channel_arand = [adapted_rand_error(_target, channel_segm)[0] for channel_segm in segm]
            per_batch_arand.append(np.min(per_channel_arand))

        # return mean arand error
        mean_arand = torch.mean(torch.tensor(per_batch_arand))
        logger.info(f"ARand: {mean_arand.item()}")
        return mean_arand

    def input_to_segm(self, input):
        """Converts input tensor to segmentation image.

        If the input is boundary pmaps, one option would be to threshold it and run connected
        components in order to return the segmentation.

        Args:
            input: 4D tensor (CDHW).

        Returns:
            Segmentation volume, 4D (segmentation per channel).
        """
        # by deafult assume that input is a segmentation volume itself
        return input


class BoundaryAdaptedRandError(AdaptedRandError):
    """
    Compute ARand between the input boundary map and target segmentation.
    Boundary map is thresholded, and connected components is run to get the predicted segmentation
    """

    def __init__(
        self,
        thresholds=None,
        use_last_target=True,
        ignore_index=None,
        input_channel=None,
        invert_pmaps=True,
        save_plots=False,
        plots_dir=".",
        **kwargs,
    ):
        super().__init__(
            use_last_target=use_last_target,
            ignore_index=ignore_index,
            save_plots=save_plots,
            plots_dir=plots_dir,
            **kwargs,
        )

        if thresholds is None:
            thresholds = [0.3, 0.4, 0.5, 0.6]
        assert isinstance(thresholds, list)
        self.thresholds = thresholds
        self.input_channel = input_channel
        self.invert_pmaps = invert_pmaps

    def input_to_segm(self, input):
        if self.input_channel is not None:
            input = np.expand_dims(input[self.input_channel], axis=0)

        segs = []
        for predictions in input:
            for th in self.thresholds:
                # threshold probability maps
                predictions_th = predictions > th

                if self.invert_pmaps:
                    # for connected component analysis we need to treat boundary signal as background
                    # assign 0-label to boundary mask
                    predictions_th = np.logical_not(predictions_th)

                predictions_th = predictions_th.astype(np.uint8)
                # run connected components on the predicted mask; consider only 1-connectivity
                seg = measure.label(predictions_th, background=0, connectivity=1)
                segs.append(seg)

        return np.stack(segs)


class GenericAdaptedRandError(AdaptedRandError):
    def __init__(
        self, input_channels, thresholds=None, use_last_target=True, ignore_index=None, invert_channels=None, **kwargs
    ):
        super().__init__(use_last_target=use_last_target, ignore_index=ignore_index, **kwargs)
        assert isinstance(input_channels, list) or isinstance(input_channels, tuple)
        self.input_channels = input_channels
        if thresholds is None:
            thresholds = [0.3, 0.4, 0.5, 0.6]
        assert isinstance(thresholds, list)
        self.thresholds = thresholds
        if invert_channels is None:
            invert_channels = []
        self.invert_channels = invert_channels

    def input_to_segm(self, input):
        # pick only the channels specified in the input_channels
        results = []
        for i in self.input_channels:
            c = input[i]
            # invert channel if necessary
            if i in self.invert_channels:
                c = 1 - c
            results.append(c)

        input = np.stack(results)

        segs = []
        for predictions in input:
            for th in self.thresholds:
                # run connected components on the predicted mask; consider only 1-connectivity
                seg = measure.label((predictions > th).astype(np.uint8), background=0, connectivity=1)
                segs.append(seg)

        return np.stack(segs)


class GenericAveragePrecision:
    def __init__(self, min_instance_size=None, use_last_target=False, metric="ap", **kwargs):
        self.min_instance_size = min_instance_size
        self.use_last_target = use_last_target
        assert metric in ["ap", "acc"]
        if metric == "ap":
            # use AveragePrecision
            self.metric = AveragePrecision()
        else:
            # use Accuracy at 0.5 IoU
            self.metric = Accuracy(iou_threshold=0.5)

    def __call__(self, input, target):
        if target.dim() == 5:
            if self.use_last_target:
                target = target[:, -1, ...]  # 4D
            else:
                # use 1st target channel
                target = target[:, 0, ...]  # 4D

        input1 = input2 = input
        multi_head = isinstance(input, tuple)
        if multi_head:
            input1, input2 = input

        input1, input2, target = convert_to_numpy(input1, input2, target)

        batch_aps = []
        i_batch = 0
        # iterate over the batch
        for inp1, inp2, tar in zip(input1, input2, target, strict=True):
            if multi_head:
                inp = (inp1, inp2)
            else:
                inp = inp1

            segs = self.input_to_seg(inp, tar)  # expects 4D
            assert segs.ndim == 4
            # convert target to seg
            tar = self.target_to_seg(tar)

            # filter small instances if necessary
            tar = self._filter_instances(tar)

            # compute average precision per channel
            segs_aps = [self.metric(self._filter_instances(seg), tar) for seg in segs]

            logger.info(f"Batch: {i_batch}. Max Average Precision for channel: {np.argmax(segs_aps)}")
            # save max AP
            batch_aps.append(np.max(segs_aps))
            i_batch += 1

        return torch.tensor(batch_aps).mean()

    def _filter_instances(self, input):
        """Filters instances smaller than 'min_instance_size' by overriding them with 0-index.

        Args:
            input: Input instance segmentation.

        Returns:
            Filtered instance segmentation.
        """
        if self.min_instance_size is not None:
            labels, counts = np.unique(input, return_counts=True)
            for label, count in zip(labels, counts, strict=True):
                if count < self.min_instance_size:
                    input[input == label] = 0
        return input

    def input_to_seg(self, input, target=None):
        raise NotImplementedError

    def target_to_seg(self, target):
        return target


class BlobsAveragePrecision(GenericAveragePrecision):
    """Computes Average Precision given foreground prediction and ground truth instance segmentation.

    Args:
        thresholds: List of thresholds for segmentation. Default: [0.4, 0.5, 0.6, 0.7, 0.8].
        metric: Metric to use ('ap' or 'acc'). Default: 'ap'.
        min_instance_size: Minimum instance size to consider.
        input_channel: Channel to use from input. Default: 0.
    """

    def __init__(self, thresholds=None, metric="ap", min_instance_size=None, input_channel=0, **kwargs):
        super().__init__(min_instance_size=min_instance_size, use_last_target=True, metric=metric)
        if thresholds is None:
            thresholds = [0.4, 0.5, 0.6, 0.7, 0.8]
        assert isinstance(thresholds, list)
        self.thresholds = thresholds
        self.input_channel = input_channel

    def input_to_seg(self, input, target=None):
        input = input[self.input_channel]
        segs = []
        for th in self.thresholds:
            # threshold and run connected components
            mask = (input > th).astype(np.uint8)
            seg = measure.label(mask, background=0, connectivity=1)
            segs.append(seg)
        return np.stack(segs)


class BlobsBoundaryAveragePrecision(GenericAveragePrecision):
    """Computes Average Precision given foreground and boundary predictions.

    Segmentation mask is computed as (P_mask - P_boundary) > th followed by a connected component.

    Args:
        thresholds: List of thresholds for segmentation. Default: [0.3, 0.4, 0.5, 0.6, 0.7].
        metric: Metric to use ('ap' or 'acc'). Default: 'ap'.
        min_instance_size: Minimum instance size to consider.
    """

    def __init__(self, thresholds=None, metric="ap", min_instance_size=None, **kwargs):
        super().__init__(min_instance_size=min_instance_size, use_last_target=True, metric=metric)
        if thresholds is None:
            thresholds = [0.3, 0.4, 0.5, 0.6, 0.7]
        assert isinstance(thresholds, list)
        self.thresholds = thresholds

    def input_to_seg(self, input, target=None):
        # input = P_mask - P_boundary
        input = input[0] - input[1]
        segs = []
        for th in self.thresholds:
            # threshold and run connected components
            mask = (input > th).astype(np.uint8)
            seg = measure.label(mask, background=0, connectivity=1)
            segs.append(seg)
        return np.stack(segs)


class BoundaryAveragePrecision(GenericAveragePrecision):
    """Computes Average Precision given boundary prediction and ground truth instance segmentation.

    Args:
        thresholds: List of thresholds for segmentation. Default: [0.3, 0.4, 0.5, 0.6].
        min_instance_size: Minimum instance size to consider.
        input_channel: Channel to use from input. Default: 0.
    """

    def __init__(self, thresholds=None, min_instance_size=None, input_channel=0, **kwargs):
        super().__init__(min_instance_size=min_instance_size, use_last_target=True)
        if thresholds is None:
            thresholds = [0.3, 0.4, 0.5, 0.6]
        assert isinstance(thresholds, list)
        self.thresholds = thresholds
        self.input_channel = input_channel

    def input_to_seg(self, input, target=None):
        input = input[self.input_channel]
        segs = []
        for th in self.thresholds:
            seg = measure.label(np.logical_not(input > th).astype(np.uint8), background=0, connectivity=1)
            segs.append(seg)
        return np.stack(segs)


class PSNR:
    """Computes Peak Signal to Noise Ratio.

    Use e.g. as an eval metric for denoising task.
    """

    def __init__(self, **kwargs):
        pass

    def __call__(self, input, target):
        input, target = convert_to_numpy(input, target)
        return peak_signal_noise_ratio(target, input)


class MSE:
    """Computes MSE between input and target."""

    def __init__(self, **kwargs):
        pass

    def __call__(self, input, target):
        input, target = convert_to_numpy(input, target)
        return mean_squared_error(input, target)


def get_evaluation_metric(config):
    """Returns the evaluation metric function based on provided configuration.

    Args:
        config: A top level configuration object containing the 'eval_metric' key.

    Returns:
        An instance of the evaluation metric.
    """

    def _metric_class(class_name):
        m = importlib.import_module("pytorch3dunet.unet3d.metrics")
        clazz = getattr(m, class_name)
        return clazz

    assert "eval_metric" in config, "Could not find evaluation metric configuration"
    metric_config = config["eval_metric"]
    metric_class = _metric_class(metric_config["name"])
    return metric_class(**metric_config)



import importlib
import os

from pytorch3dunet.unet3d.predictor import AbstractPredictor

# Fix for OpenMP library conflict on Windows
os.environ["KMP_DUPLICATE_LIB_OK"] = "TRUE"

import torch
import torch.nn as nn

from pytorch3dunet.datasets.utils import get_test_loaders
from pytorch3dunet.unet3d import utils
from pytorch3dunet.unet3d.config import TorchDevice, load_config
from pytorch3dunet.unet3d.model import get_model

logger = utils.get_logger("UNet3DPredict")


def get_predictor(model: nn.Module, config: dict) -> AbstractPredictor:
    """Create and return a predictor instance based on the configuration.

    Args:
        model: The trained model to use for prediction.
        config: Configuration dictionary containing predictor settings.

    Returns:
        A predictor instance (StandardPredictor or LazyPredictor).
    """
    output_dir = config["loaders"].get("output_dir", None)
    if output_dir is not None:
        os.makedirs(output_dir, exist_ok=True)

    predictor_config = config.get("predictor", {})
    class_name = predictor_config.get("name", "StandardPredictor")

    m = importlib.import_module("pytorch3dunet.unet3d.predictor")
    predictor_class = getattr(m, class_name)
    out_channels = config["model"].get("out_channels")
    return predictor_class(model, output_dir, out_channels, **predictor_config, device=config["device"])


def main():
    """Main entry point for prediction with 3D U-Net models.

    Loads configuration from command line arguments, creates the model, loads trained weights,
    runs predictions on test datasets, and computes evaluation metrics if specified.
    """
    # Load configuration
    config, _ = load_config()

    # Create the model
    model = get_model(config["model"])
    device = config.get("device", None)
    assert device, "Device not specified in the config file and could not be inferred automatically"
    logger.info(f"Using device: {device}")

    # Load model state
    model_path = config["model_path"]
    logger.info(f"Loading model from {model_path}...")
    utils.load_checkpoint(model_path, model)

    # use DataParallel if more than 1 GPU available
    if device == TorchDevice.CUDA and torch.cuda.device_count() > 1:
        model = nn.DataParallel(model)
        logger.info(f"Using {torch.cuda.device_count()} GPUs for prediction")
    model = model.to(device)

    logger.info("Creating predictor...")
    # create predictor instance
    predictor = get_predictor(model, config)

    metrics = []
    for test_loader in get_test_loaders(config):
        # run the model prediction on the test_loader and save the results in the output_dir
        metric = predictor(test_loader)
        if metric is not None:
            metrics.append(metric)

    if metrics:
        # average across loaders
        metrics = torch.Tensor(metrics)
        per_class_metrics = metrics.mean(dim=0)
        avg_metric = metrics.mean()
        logger.info(f"Per-class average metric: {per_class_metrics}")
        logger.info(f"Average metric: {avg_metric}")


if __name__ == "__main__":
    main()



import numpy as np

from pytorch3dunet.augment.transforms import (
    CropToFixed,
    LabelToAffinities,
    RandomGammaCorrection,
    RandomLabelToAffinities,
    Relabel,
    Transformer,
)


class TestTransforms:
    config = {"dtype": "int64"}

    def test_random_label_to_boundary(self):
        size = 20
        label = _diagonal_label_volume(size)

        transform = RandomLabelToAffinities(np.random.RandomState())
        result = transform(label)
        assert result.shape == (1,) + label.shape

    def test_random_label_to_boundary_with_ignore(self):
        size = 20
        label = _diagonal_label_volume(size, init=-1)

        transform = RandomLabelToAffinities(np.random.RandomState(), ignore_index=-1)
        result = transform(label)
        assert result.shape == (1,) + label.shape
        assert -1 in np.unique(result)

    def test_label_to_boundary(self):
        size = 20
        label = _diagonal_label_volume(size)

        # this transform will produce 2 channels
        transform = LabelToAffinities(offsets=(2, 4), aggregate_affinities=True)
        result = transform(label)
        assert result.shape == (2,) + label.shape
        assert np.array_equal(np.unique(result), [0, 1])

    def test_label_to_boundary_with_ignore(self):
        size = 20
        label = _diagonal_label_volume(size, init=-1)

        transform = LabelToAffinities(offsets=(2, 4), ignore_index=-1, aggregate_affinities=True)
        result = transform(label)
        assert result.shape == (2,) + label.shape
        assert np.array_equal(np.unique(result), [-1, 0, 1])

    def test_label_to_boundary_no_aggregate(self):
        size = 20
        label = _diagonal_label_volume(size)

        # this transform will produce 6 channels
        transform = LabelToAffinities(offsets=(2, 4), aggregate_affinities=False)
        result = transform(label)
        assert result.shape == (6,) + label.shape
        assert np.array_equal(np.unique(result), [0, 1])

    def test_relabel(self):
        label = np.array([[10, 10, 10], [0, 0, 0], [5, 5, 5]])
        r = Relabel()
        result = r(label)
        assert np.array_equal(result, np.array([[1, 1, 1], [0, 0, 0], [2, 2, 2]]))

    def test_BaseTransformer(self):
        config = {
            "raw": [{"name": "Standardize"}, {"name": "ToTensor", "expand_dims": True}],
            "label": [{"name": "ToTensor", "expand_dims": False, "dtype": "int64"}],
            "weight": [{"name": "ToTensor", "expand_dims": False}],
        }
        base_config = {"mean": 0, "std": 1}
        transformer = Transformer(config, base_config)
        raw_transforms = transformer.raw_transform().transforms
        assert raw_transforms[0].mean == 0
        assert raw_transforms[0].std == 1
        assert raw_transforms[1].expand_dims
        label_transforms = transformer.label_transform().transforms
        assert not label_transforms[0].expand_dims
        assert label_transforms[0].dtype == "int64"

    def test_StandardTransformer(self):
        config = {
            "raw": [
                {"name": "Standardize"},
                {"name": "RandomContrast", "execution_probability": 0.5},
                {"name": "RandomFlip"},
                {"name": "RandomRotate90"},
                {"name": "ToTensor", "expand_dims": True},
            ],
            "label": [
                {"name": "RandomFlip"},
                {"name": "RandomRotate90"},
                {"name": "ToTensor", "expand_dims": False, "dtype": "int64"},
            ],
        }
        base_config = {"mean": 0, "std": 1}
        transformer = Transformer(config, base_config)
        raw_transforms = transformer.raw_transform().transforms
        assert raw_transforms[0].mean == 0
        assert raw_transforms[0].std == 1
        assert raw_transforms[1].execution_probability == 0.5
        assert raw_transforms[4].expand_dims
        label_transforms = transformer.label_transform().transforms
        assert len(label_transforms) == 3

    def test_AnisotropicRotationTransformer(self):
        config = {
            "raw": [
                {"name": "Standardize"},
                {"name": "RandomContrast", "execution_probability": 0.5},
                {"name": "RandomFlip"},
                {"name": "RandomRotate90"},
                {"name": "RandomRotate", "angle_spectrum": 17, "axes": [[2, 1]]},
                {"name": "ToTensor", "expand_dims": True},
            ],
            "label": [
                {"name": "RandomFlip"},
                {"name": "RandomRotate90"},
                {"name": "RandomRotate", "angle_spectrum": 17, "axes": [[2, 1]]},
                {"name": "ToTensor", "expand_dims": False, "dtype": "int64"},
            ],
        }
        base_config = {"mean": 0, "std": 1}
        transformer = Transformer(config, base_config)
        raw_transforms = transformer.raw_transform().transforms
        assert raw_transforms[0].mean == 0
        assert raw_transforms[0].std == 1
        assert raw_transforms[1].execution_probability == 0.5
        assert raw_transforms[4].angle_spectrum == 17
        assert raw_transforms[4].axes == [[2, 1]]
        label_transforms = transformer.label_transform().transforms
        assert len(label_transforms) == 4

    def test_LabelToBoundaryTransformer(self):
        config = {
            "raw": [
                {"name": "Standardize"},
                {"name": "RandomContrast", "execution_probability": 0.5},
                {"name": "RandomFlip"},
                {"name": "RandomRotate90"},
                {"name": "RandomRotate", "angle_spectrum": 17, "axes": [[2, 1]], "mode": "reflect"},
                {"name": "ToTensor", "expand_dims": True},
            ],
            "label": [
                {"name": "RandomFlip"},
                {"name": "RandomRotate90"},
                {"name": "RandomRotate", "angle_spectrum": 17, "axes": [[2, 1]], "mode": "reflect"},
                {"name": "LabelToAffinities", "offsets": [2, 4, 6, 8]},
                {"name": "ToTensor", "expand_dims": False, "dtype": "int64"},
            ],
        }
        base_config = {"mean": 0, "std": 1}
        transformer = Transformer(config, base_config)
        raw_transforms = transformer.raw_transform().transforms
        assert raw_transforms[0].mean == 0
        assert raw_transforms[0].std == 1
        assert raw_transforms[1].execution_probability == 0.5
        assert raw_transforms[4].angle_spectrum == 17
        assert raw_transforms[4].axes == [[2, 1]]
        assert raw_transforms[4].mode == "reflect"
        label_transforms = transformer.label_transform().transforms
        assert label_transforms[2].angle_spectrum == 17
        assert label_transforms[2].axes == [[2, 1]]
        assert label_transforms[2].mode == "reflect"
        # 3 conv kernels per offset
        assert len(label_transforms[3].kernels) == 12

    def test_RandomLabelToBoundaryTransformer(self):
        config = {
            "raw": [
                {"name": "Normalize"},
                {"name": "RandomContrast", "execution_probability": 0.5},
                {"name": "RandomFlip"},
                {"name": "RandomRotate90"},
                {"name": "RandomRotate", "angle_spectrum": 17, "axes": [[2, 1]], "mode": "reflect"},
                {"name": "ToTensor", "expand_dims": True},
            ],
            "label": [
                {"name": "RandomFlip"},
                {"name": "RandomRotate90"},
                {"name": "RandomRotate", "angle_spectrum": 17, "axes": [[2, 1]], "mode": "reflect"},
                {"name": "RandomLabelToAffinities", "max_offset": 4},
                {"name": "ToTensor", "expand_dims": False, "dtype": "int64"},
            ],
        }
        base_config = {"mean": 0, "std": 1}
        transformer = Transformer(config, base_config)
        label_transforms = transformer.label_transform().transforms
        assert label_transforms[3].offsets == (1, 2, 3, 4)

    def test_crop_to_fixed_when_crop_bigger_than_volume(self):
        m = np.random.rand(200, 200, 200)

        rs1 = np.random.RandomState(13)

        t = CropToFixed(rs1, size=(256, 256))

        expected = np.pad(m, ((0, 0), (28, 28), (28, 28)), mode="reflect")

        assert np.array_equal(expected, t(m))

    def test_crop_to_fixed(self):
        m = np.random.rand(200, 200, 200)

        rs1 = np.random.RandomState(13)
        rs2 = np.random.RandomState(13)

        t = CropToFixed(rs1, size=(128, 128))

        r = 200 - 128
        y_start = rs2.randint(r)
        x_start = rs2.randint(r)

        m_crop = m[:, y_start : y_start + 128, x_start : x_start + 128]

        assert np.array_equal(m_crop, t(m))

    def test_random_gamma_correction(self):
        m = np.random.rand(200, 200, 200)
        t = RandomGammaCorrection(np.random.RandomState(), gamma=(1.1, 2.0), execution_probability=1.0)
        # Output is darker for gamma > 1
        assert np.mean(m) > np.mean(t(m))


def _diagonal_label_volume(size, init=1):
    label = init * np.ones((size, size, size), dtype=np.int32)
    for i in range(size):
        for j in range(size):
            for k in range(size):
                if i + j > 2 * k:
                    label[i, j, k] = 3
    return label
